dojo.provide('ibm_ilog.diagram.Port');

dojo.require('ibm_ilog.diagram.util.ErrorReporter');
dojo.require('ibm_ilog.diagram.util.ClippingUtil');
dojo.require("ibm_ilog.diagram.util.Geometry");

dojo.require('dojox.gfx');

dojo.declare("ibm_ilog.diagram.PortBase", null, {
	// summary:
	//		The PortBase class represents the base class for ports.
	// description:
	//		A port defines a connection point between a Link and a Node.
	//		The PortBase class is intended as the base class for subclasses that
	//		define different types of ports. PortBase instances should not be created
	//		directly, only its subclasses should be used.
	
	_owner: null,
	_links: null,
	_clipOnShape: null,
	
	getOwner: function() {
		// summary:
		//		Gets the owner Node of this port.
		// returns:
		//		The Node to which this port belongs.
		return this._owner;
	},
	
	_setOwner: function(owner) {
		// summary:
		//		Sets the owner Node of this port.
		// owner: ibm_ilog.diagram.Node:
		//		The new owner of this port.
		// description:
		//		This method is internal and should not be called directly by application code.
		
		if (owner != this._owner) {
			this._owner = owner;
			this.invalidateLinks();
		}
	},
	
	getConnectionPoint: function(link, parent, referencePoint, connectionPoint, originPoint, start){
		// summary: 
		//		Gets the connection point and the origin point for this port.
		// link: ibm_ilog.diagram.Link:
		//		The Link that is (or will be) connected to this port.
		// referencePoint:
		//		The opposite point of the segment connected to this port.
		// parent:
		//		The parent object (usually a Graph) that determines the coordinate space
		//		the connection point is relative to.
		// connectionPoint:
		//		An object (with 'x' and 'y' properties) into which the connection point will be stored.
		//		The connection point is the point where the link will start or end.
		// originPoint:
		//		An object (with 'x' and 'y' properties) into which the origin point will be stored.
		//		The origin point is the point towards which the link will 'aim' (typically the center of the node),
		//		and is used to determine the shape of orthogonal links for example.
		// returns:
		//		The 'connectionPoint' argument.
		// description:
		//		This is the central method of the PortBase class and is called by the Link object
		//		to get its start and end points. This method calls computeConnectionPoint, which is
		//		the method that actually calculates the connection point in the PortBase subclasses.
		//		Then, getConnectionPoint applies the clipping if getClipOnShape() is true.
		//		So, getConnectionPoint should not be overridden in subclasses, computeConnectionPoint
		//		is the method that is overridden instead.
		this.computeConnectionPoint(link, parent, referencePoint, connectionPoint, originPoint, start);
		if(this.getClipOnShape()){
			var p = this._clipToShape(link,parent,referencePoint, connectionPoint, originPoint, start);
			connectionPoint.x = p.x;
			connectionPoint.y = p.y;
		}
		return connectionPoint;
	},	
	
	computeConnectionPoint: function(link, parent, referencePoint, connectionPoint, originPoint, start){
		// summary: 
		//		Computes the connection point and the origin point for a given subclass of PortBase.
		// link: ibm_ilog.diagram.Link:
		//		The Link that is (or will be) connected to this port.
		// referencePoint:
		//		The opposite point of the segment connected to this port.
		// parent:
		//		The parent object (usually a Graph) that determines the coordinate space
		//		the connection point is relative to.
		// connectionPoint:
		//		An object (with 'x' and 'y' properties) into which the connection point will be stored.
		//		The connection point is the point where the link will start or end.
		// originPoint:
		//		An object (with 'x' and 'y' properties) into which the origin point will be stored.
		//		The origin point is the point towards which the link will 'aim' (typically the center of the node),
		//		and is used to determine the shape of orthogonal links for example.
		// returns:
		//		The 'connectionPoint' argument.
		// description:
		//		This method is called by PortBase.getConnectionPoint, and should be overridden in
		//		PortBase subclasses to actually implement the calculation of the connection point.
		//		Clipping on the node's shape is implemented by getConnectionPoint, so
		//		computeConnectionPoint does not have to care about it.
		//		The base implementation for the PortBase class returns the center of the node for both
		//		the connection point and the origin point.
		
		var bounds = this.getConnectionBounds(parent);
		connectionPoint.x = originPoint.x = bounds.x + bounds.width / 2;
		connectionPoint.y = originPoint.y = bounds.y + bounds.height / 2;
		return connectionPoint;
	},
	
	getConnectionBounds: function(parent) {
		// summary:
		//		Gets the bounds of the owner node's base shape used to compute the connection point.
		// parent:
		//		The parent object (usually a Graph) that determines the coordinate space
		//		the bounds are relative to.
		// returns:
		//		A rectangle (with 'x', 'y', 'width' and 'height' properties) containing the bounds
		//		used to compute the connection point.
		// description:
		//		This method returns the bounds that will be used by the port to compute the connection and origin points.
		//		The default implementation returns the bounds of the base shape returned by getBaseShape().
		//		This method	can be overridden to return different bounds.
		
		if(!parent)
			parent = this._owner.getParent();
		
		var baseShape = this.getBaseShape();
        var rect = baseShape.getBoundingBox();
        var m = dojox.gfx.matrix;
        // t1: transform from surface -> base shape
        var t1 = baseShape._getRealMatrix() || m.identity;
        // t2: transform from surface -> parent
        var t2 = parent._getRealMatrix() || m.identity;
        var t = m.multiply(t2.inverse(), t1);
		rect = t.transformRectangle(rect);
        return rect;
	},
	
	needsReferencePoint: function() {
		// summary:
		//		Indicates whether or not this port needs a reference point to compute the connection point.
		// returns:
		//		True if this port needs a reference point to compute the connection point,
		//		or false otherwise.
		// description:
		//		This method is used by a PortBase subclass implementation to signal whether or not
		//		it needs (and uses) the referencePoint argument passed to the getConnectionPoint method.
		//		A subclass must return true if it actually uses the referencePoint, or false if it does
		//		not use it. If needsReferencePoint returns true, the caller of the getConnectionPoint method
		//		(usually the Link object) must pass a valid referencePoint. If needsReferencePoint returns false,
		//		the PortBase subclass guarantees that the referencePoint will never be used or referenced,
		//		so the caller may pass an invalid point, or a null or undefined object.
		//		By default the method returns true if getClipOnShape() returns true, of false otherwise.
		//		PortBase subclasses that always need a reference point must override this method and return true.
		
		return this._clipOnShape || false;
	},
	
	invalidateLinks: function() {
		// summary:
		//		Invalidates the shape of all the links connected to this port. 
		// description:
		//		Invalidates the shape of all the links connected to this port. 
		//		This method must be called by PortBase subclasses when a change of the port
		//		causes the connection point to change.
		if(this._links)
			for(var i = 0; i < this._links.length; i++)
				this._links[i].invalidateLinkShape();
				
	},
	
	_addLink: function(link) {
		// summary:
		//		Adds a link to the list of links connected to this port.
		// description:
		//		This method is internal and should not be called directly by application code.
		if(!this._links)
			this._links = [ link ];
		else
			this._links.push(link);
	},
	
	_removeLink: function(link) {
		// summary:
		//		Removes a link from the list of links connected to this port.
		// description:
		//		This method is internal and should not be called directly by application code.
		if (this._links) {
			var index = dojo.indexOf(this._links, link);
			if (index >= 0) {
				if(this._links.length == 1)
					this._links = null;
				else
					this._links.splice(index, 1);
			}
		}
	},
	
	getLinks: function() {
		// summary:
		//		Gets the links connected to this port.
		// returns:
		//		An array of Links connected to this port.
		// description:
		//		Note that this method always returns a newly allocated array.
		var links = [];
		if (this._links) {
			for(var i = 0; i < this._links.length; i++)
			links.push(this._links[i]);
		}
		return links;
	},
	
	getLinksCount: function() {
		// summary:
		//		Gets the number of links connected to this port.
		
		return this._links ? this._links.length : 0;
	},

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////   CLIPPING TOOLS ///////////////////////////////////////////////////////////////////	
/////////////////////////////////////////////////////////////////////////////////////////////////////////////

	setClipOnShape: function(clipOnShape){
		// summary:
		//		Determines whether the connection point is "clipped" on the base shape of the node.
		// description:
		//		If clipOnShape is true, the getConnectionPoint method will extend the segment defined
		//		by the reference point and the connection point so that the connection point is on the
		//		outline of the node's shape, so that the connection point always "touches" the node shape
		//		even if the node is not rectangular (for example, if the node has an elliptical of polygonal shape).
		//		If clipOnShape is false, the getConnectionPoint always returns the return of computeConnectionPoint
		//		without modifying it, and the point may not be exactly on the boundary on the node shape.
		// clipOnShape:
		//		If true, the connection point is clipped on the node's shape, otherwise the connection point
		//		is the point returned by computeConnectionPoint.
		
		if (clipOnShape != this._clipOnShape) {
			this._clipOnShape = clipOnShape;
			this.invalidateLinks();
		}
	},
	
	getClipOnShape: function(){
		// summary:
		//		Gets the flag that determines whether the connection point is "clipped" on the base shape of the node.
		// description:
		//		If clipOnShape is true, the getConnectionPoint method will extend the segment defined
		//		by the reference point and the connection point so that the connection point is on the
		//		outline of the node's shape, so that the connection point always "touches" the node shape
		//		even if the node is not rectangular (for example, if the node has an elliptical of polygonal shape).
		//		If clipOnShape is false, the getConnectionPoint always returns the return of computeConnectionPoint
		//		without modifying it, and the point may not be exactly on the boundary on the node shape.
		// returns:
		//		True if the connection point is clipped on the node's shape, or false otherwise.
		
		return this._clipOnShape;
	},
	
	getBaseShape : function() {
		// summary:
		//		Gets the shape contained in the owner node's template to which links should be connected.
		// description:
		//		The base shape is used by getConnectionBounds to compute the default connection bounds,
		//		and also when clipOnShape is true to clip the connection point if the base shape is
		//		non-rectangular.
		//		By default the base shape is determined by calling the getBaseShape() of the owner node,
		//		but this can be overridden to connect links to a different template part of the node.
		
		return this._owner.getBaseShape();
	},
	
	_clipToShape: function(link,parent,referencePoint, connectionPoint, originPoint, start){
		
		var node = this._owner;
		var shape = this.getBaseShape();
		var bounds = this.getConnectionBounds(parent);
		
		if(shape == node){
			var shape = ibm_ilog.diagram.util.Geometry.Rect(bounds.x,bounds.y,bounds.width,bounds.height);
		}
			
		var resClipPoint = ibm_ilog.diagram.util.ClippingUtil.clipToShape(node,parent,shape,bounds,originPoint, connectionPoint, referencePoint)

		return resClipPoint?resClipPoint:connectionPoint;
	}
		
});

dojo.declare("ibm_ilog.diagram.BasicPort", ibm_ilog.diagram.PortBase, {
	// summary:
	//		The BasicPort class represents ports that define fixed connection points
	//		relative to the rectangular bounds of a node.
	//		The 'position' property is an object with 'x' and 'y' properties representing
	//		ratios (between 0 and 1) of the width and height of the node. For example,
	//		the position { x: 0.5, y: 0.5} defines a connection point at the center of a node, and
	//		the position { x: 1, y: 0 } defines a connection point at the top-right of a node.
	//		The 'offset' property (another x/y object, but in pixels this time) can be used
	//		to alter slightly this relative position.
	
	// Allows to test the class faster that testing declaredClass
	_isIBMDiagramBasicPort: true,
	
	_position: null,
	_offset: null,
	_movable: true,
	
    constructor: function() {
		// summary:
		//		Creates a new BasicPort with a default position set to the center of the node.
		// description:
		//		The default position is { x: 0.5, y: 0.5 }, and the default offset is { x: 0, y: 0 }.
		this._position = { x: 0.5, y: 0.5 };
		this._offset = { x: 0, y: 0 };
	},
	
	getPosition: function() {
		// summary:
		//		Gets the position of this port.
		// returns:
		//		An object with 'x' and 'y' properties representing a ratio (between 0 and 1) of the node width/height.
		return this._position;
	},
	
	setPosition: function(position) {
		// summary:
		//		Sets the position of this port.
		// position:
		//		An object with 'x' and 'y' properties representing a ratio (between 0 and 1) of the node width/height.
		
		if(typeof(position) != 'object' || !('x' in position) || !('y' in position))
			ibm_ilog.diagram.util.ErrorReporter.error("UnexpectedArgument",this, "setPosition", "position", position);
		
		if (position.x != this._position.x ||
		    position.y != this._position.y) {
			this._position.x = position.x;
			this._position.y = position.y;
			this.invalidateLinks();
		}
	},
		
    isMovable: function() {
        // summary:
        //		Gets the movable state of this port.
        // returns:
        //		True if graph layout and user interaction are allowed to change the position 
	    //      of the port.
        //		False if graph layout and user interaction are not allowed to change the 
		//      position of the port.
        
		// AVA: the design of the class hierarchy for Ports is different in  
		// Diagram for Flex in the sense that the base class (PortBase) defines
		// a "move" function (default implementation is empty). Hence, the functions
		// "set/isMovable" plus the marker "supportsMove" are also defined at the
		// base class. This allows the code that uses the port to treat all ports in
		// a generic way, without the need to check the port class.
		// Here, the function which allows to move the port ("setPosition") exists
		// only on BasicPort, not on PortBase. Therefore, we cannot place "set/isMovable"
		// and "supportsMove" at PortBase. Not a big deal, but maybe it would be worth
		// to review the design.
		// With the current design, Diagram.flex' "supportsMove" makes no sense here,
		// therefore I only added "set/isMovable".  
        return this._movable ? true : false;
    },
    
    setMovable: function(m) {
        // summary:
        //		Sets the movable state of this port. The ports are movable by default.
        // m: Boolean:
        //		If 'm' is true, graph layout and user interaction are allowed to change the position 
	    //      of the port.
        //		If 'm' is false, graph layout and user interaction are not allowed to change the 
		//      position of the port.
        
        this._movable = m;
    },
	
	getOffset: function() {
		// summary:
		//		Gets the offset of this port.
		// returns:
		//		An object with 'x' and 'y' properties representing horizontal/vertical offsets in pixels.
		return this._offset;
	},
	
	setOffset: function(offset) {
		// summary:
		//		Sets the offset of this port.
		// offset:
		//		An object with 'x' and 'y' properties representing horizontal/vertical offsets in pixels.
		
		if(typeof(offset) != 'object' || !('x' in offset) || !('y' in offset))
			ibm_ilog.diagram.util.ErrorReporter.error("UnexpectedArgument",this, "setOffset", "offset", offset);
		
		if (offset.x != this._offset.x ||
		    offset.y != this._offset.y) {
			this._offset.x = offset.x;
			this._offset.y = offset.y;
			this.invalidateLinks();
		}
	},
	
	computeConnectionPoint: function(link, parent, referencePoint, connectionPoint, originPoint, start){
		// summary:
		//		Implements the calculation of the connection and origin points for the BasicPort class.
		// description:
		//		The connection point is computed based on the position and offset properties.
		//		The origin point is always the center of the node.
		
		var bounds = this.getConnectionBounds(parent);
		connectionPoint.x = bounds.x + bounds.width * this._position.x;
		connectionPoint.y = bounds.y + bounds.height * this._position.y;
		originPoint.x = bounds.x + bounds.width/2;
		originPoint.y = bounds.y + bounds.height/2;
		connectionPoint.x += this._offset.x;
		connectionPoint.y += this._offset.y;
		return connectionPoint;
	}
});

dojo.declare("ibm_ilog.diagram.AutomaticPort", ibm_ilog.diagram.PortBase, {
	// summary:
	//		The AutomaticPort class represents a port that connects a link to the middle of the
	//		closest side of a node.
	
	needsReferencePoint: function(){
		// summary:
		//		Returns true to indicate that the automatic port needs a reference point
		//		to compute the connection point.
		return true;
	},
		
	computeConnectionPoint: function(link, parent, referencePoint, connectionPoint, originPoint, start){
		// summary:
		//		Implements the calculation of the connection and origin points for the AutomaticPort class.
		// description:
		//		Computes the connection point as the middle of the side of the node that is closest
		//		to the reference point.
		//		The origin point is always the center of the node.
		
		var bounds = this.getConnectionBounds(parent);
		var cx = bounds.x + bounds.width/2;
		var cy = bounds.y + bounds.height/2;
		originPoint.x = cx;
		originPoint.y = cy;
		connectionPoint.x = cx;
		connectionPoint.y = cy;
		if (bounds.width > 0 && bounds.height > 0) {
			var dx = referencePoint.x - cx;
			var dy = referencePoint.y - cy;
			var refRatio = dx != 0 ? (dy > 0 ? dy : -dy) / (dx > 0 ? dx : -dx) : 2;
			if (refRatio > 1) {
				// above or below: determine which from sign of dy
				connectionPoint.y = bounds.y;
				if (dy > 0) 
					connectionPoint.y += bounds.height;
			}
			else {
				// left or right: determine which from sign of dx
				connectionPoint.x = bounds.x;
				if (dx > 0) 
					connectionPoint.x += bounds.width;
			}
		}
		return connectionPoint;
	}
});
