dojo.provide('ibm_ilog.diagram.gfxwidget.GfxWidget');

dojo.require('dojox.gfx');
dojo.require('dojox.gfx.shape');
dojo.require('dojox.gfx.matrix');
dojo.require('ibm_ilog.diagram.base');
dojo.require('ibm_ilog.diagram.util.Batch');
dojo.require('dijit._base');


(function () {

	var __set = ibm_ilog.diagram._base.dojoext.SetAttrName;
	
dojo.declare('ibm_ilog.diagram.gfxwidget.GfxWidget', [dojox.gfx.Rect], {
	// summary:
	//		A custom GFX shape that allows to put a Dijit widget in a Dojo Diagrammer node or link.
	// description:
	//		The <code>GfxWidget</code> class is a custom GFX shape that is intended to be used in
	//		Dojo Diagrammer node and link templates to integrate Dijit widgets in nodes or links.
	//		The widget type can be specified either by specifying the <code>dojoType</code> property
	//		of the shape, or by passing a function as the <code>createWidget</code> property of the shape.
	//		<p>
	//		Note that this special shape has a number of limitations:
	//		<ul>
	//		<li>The widget does not zoom when the graph is zoomed (only its position will be updated).</li>
	//		<li>The stacking order of widgets may be inconsistent with the order of their parent group shapes.</li>
	//		<li>If the shape is used outside of a Diagram widget, the HTML element that contains the surface
	//		must have its <code>position</code> style set to <code>absolute</code>.</li>
	//		</ul>
	//		</p>
	//	example:
	//		See the widgets/widgets.html sample for examples.
	
	_widget:null,
	
	constructor: function(){
		// summary:
		//		Creates an empty GfxWidget instance.
		// description:
		//		You must call <code>setShape</code> to specify the widget class and attributes.
		
		this.shape = dojo.clone(ibm_ilog.diagram.gfxwidget.GfxWidget.defaultWidget);
		this.bbox = null;
	},
	
	setShape: function(shape){
		// summary:
		//		Sets the attributes of the GfxWidget shape.
		// description:
		//		The <code>shape</code> parameter is an object with the following properties:
		//		<table style="border-collapse:collapse;" border="1">
		//			<tr>
		//				<th><b>Property</b></th>	<th><b>Description</b></th>
		//			</tr>
		//			<tr>
		//				<td>type</td>				<td>must be <code>'widget'</code></td>
		//			</tr>
		//			<tr>
		//				<td>dojoType</td>			<td>The type of the widget, for example <code>'dijit.form.Button'</code>. If <code>createWidget</code> is specified, dojoType is ignored.</td>
		//			</tr>
		//			<tr>
		//				<td>createWidget</td>		<td>A function that is called to create the widget. The function will be called with the <code>GfxWidget</code> instance as argument.</td>
		//			</tr>
		//			<tr>
		//				<td>x,y</td>				<td>The position of the widget.</td>
		//			</tr>
		//			<tr>
		//				<td>width,height</td>		<td>The size of the shape. If no size is specified, the size of the widget is used.</td>
		//			</tr>
		//			<tr>
		//				<td><em>widget attribute</em></td>		<td>Any other property is passed to the widget using an <code>attr()</code> call. For example, <code>label: 'Click Here'</code> to set the label of a <code>Button</code>.</td>
		//			</tr>
		//		</table>
		// shape: Object
		//		A hash that contains the properties of the widget.
		
		var dflt = ibm_ilog.diagram.gfxwidget.GfxWidget.defaultWidget;
		
		shape = dojo.mixin(dojo.clone(dflt), shape);
		
		if (shape) {
			if (!this._widget) {
				var w = null;
				
				if (typeof(shape.createWidget) == 'function') {
					w = shape.createWidget.call(null, this);
				} else if (shape.dojoType) {
					dojo["require"](shape.dojoType);
					var c = dojo.getObject(shape.dojoType);
					w = new c(shape);
				}
				
				if(w){
					this._widget = w;
					w._gfxShape = this;
					
					w.startup();
					
					var n = w.domNode;
					
					n.style.position = "absolute";
				}
			}
			for (var p in shape) {
				if (!(p in dflt)) {
					this._widget[__set](p, shape[p]);
				}
			}
		}
		
		this.shape = shape;
		this.bbox = null;
		
		return this;
	},
	
	prepare: function(container) {
		//this._container = container;
		//this._container.appendChild(this._widget.domNode);
	},
	
	_setParent: function(parent){
		var ret = this.inherited(arguments);
		
		for (var p = this.getParent(); p != null; p = p.getParent ? p.getParent() : null) {
			if(p instanceof dojox.gfx.Surface){
				p._parent.appendChild(this._widget.domNode);
				break;
			} else {
				dojo.connect(p, "setTransform", this, this._updateWidgetPosition);
				if(p._isIBMDiagramLink)
					dojo.connect(p, "validateLinkShape", this, this._updateWidgetPosition);
			}
		}
		
		this._updateWidgetPosition();
		
		return ret;
	},


	_minHeight: 0,
	
	_minWidth: 0,

	_minFontPxSize: 0,

	setWidth: function(w) {
		if(this._minWidth && w<this._minWidth) {
			w = this._minWidth;
		}
		dojo.style(this._widget.domNode,"width",w+"px");		
	},

	setHeight: function(h) {
		if(this._minHeight && h<this._minHeight) {
			h = this._minHeight;
		}
		dojo.style(this._widget.domNode,"minHeight",h+"px"); // hack for textarea
		dojo.style(this._widget.domNode,"height",h+"px");
	},

	setMinWidth: function(w) {
		this._minWidth = w;
		if(w && this.getWidth()<w) {
			this.setWidth(w);
		}
	},

	setMinHeight: function(h) {
		this._minHeight = h;
		if(h && this.getHeight()<h) {
			this.setHeight(h);
		}
	},

	getMinWidth: function() {
		return this._minWidth || 0;
	},

	getMinHeight: function() {
		return this._minHeight || 0;
	},
	
	setFontPxSize: function(sz) {
		if(this._minFontPxSize && sz<this._minFontPxSize) {
			sz = this._minFontPxSize;
		}
		dojo.style(this._widget.domNode,"fontSize",sz+"px");
	},
	
	setFontMinPxSize: function(sz) {
		this._minFontPxSize = sz;
		if(sz && this.getFontPxSize()<sz) {
			this.setFontPxSize(sz);
		}
	},


	getFontMinPxSize: function() {
		return this._minFontPxSize || 0;
	},

	getWidth: function() {
		return dojox.gfx.normalizedLength(dojo.style(this._widget.domNode,"width"));
	},

	getHeight: function() {
		return dojox.gfx.normalizedLength(dojo.style(this._widget.domNode,"height"));
	},

	getFontPxSize: function() {
		return dojox.gfx.pt2px(parseFloat(dojo.style(this._widget.domNode,"font-size")));
	},	

	setTransform: function(t){
		// summary:
		//		Overridden from the base GFX shape class.
		// tags:
		//		private
		
		this.inherited(arguments);
		this._updateWidgetPosition();
		return this;
	},
	
	getBoundingBox: function(){
		// summary:
		//		Overridden from the base GFX shape class.
		// tags:
		//		private
		
		if (!this.bbox) {
			if (this.shape.width && this.shape.height) {
				this.bbox = {
					x: this.shape.x,
					y: this.shape.y,
					width: this.shape.width,
					height: this.shape.height
				};
			}
			else {
				var rect = this._widget.domNode.getBoundingClientRect();
				this.bbox = {
					x: this.shape.x,
					y: this.shape.y,
					width: rect.right - rect.left,
					height: rect.bottom - rect.top
				};
			}
		}
		return this.bbox;
	},
	
	_updateWidgetPosition: function(){
		var box = this.getBoundingBox();
		var t = this._getRealMatrix();
		if(t)
			box = t.transformRectangle(box);
		this._widget.domNode.style.left = box.x + "px";
		this._widget.domNode.style.top = box.y + "px";
	}
});

ibm_ilog.diagram.gfxwidget.GfxWidget._nextId = 1;
ibm_ilog.diagram.gfxwidget.GfxWidget.defaultWidget = { type:'widget', x:0, y:0, width:0, height:0, r:0, dojoType:'', createWidget:'' };
ibm_ilog.diagram.gfxwidget.GfxWidget.nodeType = dojox.gfx.Rect.nodeType;

ibm_ilog.diagram.gfxwidget.GfxWidget.byWidget = function(widget){
	// summary:
	//		Returns the GfxWidget instance that created the specified widget.
	
	return widget._gfxShape || null;
};

var g = dojox.gfx;

//
// Extends createShape to support our own shape types
//

function extendCreateShape(shapeType) {
    var method = "createShape";
    var old = shapeType.prototype[method];
    shapeType.prototype[method] = function() {
        var ret = old.apply(this, arguments);
        if (!ret) {
            var sh = arguments[0];
            switch (sh.type) {
                case 'widget':
                    ret = this.createObject(ibm_ilog.diagram.gfxwidget.GfxWidget, sh);
                    break;
            }
        }
        return ret;
    };
}
extendCreateShape(g.Surface);
extendCreateShape(g.Group);

})();
