dojo.provide('ibm_ilog.diagram.graphlayout.WebWorkerLayout');
dojo.require('ibm_ilog.diagram.graphlayout.ServerSideLayout');

dojo.declare('ibm_ilog.diagram.graphlayout.WebWorkerLayout', ibm_ilog.diagram.graphlayout._AsynchronousLayout, 
{
	// summary:
	//		Executes a graph layout asynchronously in a Web Worker.
	// description:
	//		The WebWorkerLayout class is similar to the ServerSideLayout class,
	//		but instead of sending a request to a REST service on a remote server,
	//		it executes the graph layout on the client asynchronously using an HTML 5 Web Worker.
	//		This class can be used only if the browser supports web workers (which is not
	//		the case of Internet Explorer, for example).
	//		
	//		The main advantage of executing a graph layout in a Web Worker is that the browser UI
	//		is not blocked while the layout is executed. On the other hand, executing the
	//		layout in a worker thread has a cost, mainly due to the initialization of the
	//		worker which must load its own version of Dojo and the graph layout code before the layout
	//		can be executed. In addition, the graph topology is sent to the web worker and
	//		returned back to the calling thread by serializing the whole graph as a JSON string,
	//		which takes some additional time for big graphs.
	//		
	//		A WebWorkerLayout object is constructed with a Graph object.
	//		To execute a layout, the user must select the algorithm with the standard 
	//		client-side GraphLayout API:
	//		
	//		graph.setNodeLayout(x);
	//		graph.setLinkLayout(y);
	//		
	//		After this, the WebWorkerLayout object can be used to execute these 
	//		algorithms in a web worker by calling:
	//		
	//		webWorkerLayout.layout();
	//		
	//		This will obtain the currently configured layouts on the graph object,
	//		including optional settings for laying out internal subgraphs, gather
	//		their configured parameters and requesting the equivalent layout in the
	//		web worker.
	//
	//		Each time a layout is requested, the graph and the layout settings are
	//		serialized into a JSON object sent in the request.
	//
	//		The layout() method returns a dojo.Deferred which is signaled after the
	//		execution is complete.
	//		
	//		The web worker is reused if possible when the same WebWorkerLayout object
	//		is called several times on the same graph. The first layout execution will take more time
	//		due to the fact that the worker must load Dojo and the graph layout code.
	//		Subsequent layout executions will be faster.
	//		
	//		The WebWorkerLayout class cannot be used if Dojo is loaded from a CDN,
	//		so you must have a local copy of Dojo on your web site.
	
	_worker: null,
	
	_debugPrefix: "WebWorkerLayout: ",
	
	// dojoDir: String
	//		The parent directory of dojo/, relative to the ibm_ilog/diagram/graphlayout/webworker/ directory.
	//		This will be passed to the Web Worker thread to load Dojo.
	//		If not specified, this will be computed from the document URL and the dojo.baseUrl.
	dojoDir: null,
	
	_start: function(data) {
		//
		//	summary:
		//		Creates the Worker object to execute the layout, and applies the results.
		//	returns:
		//		a dojo.Deferred to signal when the process finishes.
		//

		// the worker is reused if already created
		if(!this._worker){
			// create and setup worker
			// first, compute the relative URL from the main application to the _Worker.js:
			this._worker = new Worker(dojo.moduleUrl('ibm_ilog.diagram', 'graphlayout/webworker/_Worker.js'));
			this._worker.onmessage = dojo.hitch(this, function(e){
				
				/* DEBUG To active it for _GraphAdapter, see the DEBUG code in _GraphAdapter.
				 if (e.data.type == "debug") {
            console.log(e.data.message);
            return;
 				 } */
 				 
				var data = dojo.fromJson(e.data);
				if (data._error) {
					this._onError(data);
					this._worker = null; // don't reuse worker after an error...
				} else {
					this._laidout = data;
					this._preLayoutDeferred.addCallback(dojo.hitch(this,"_onComplete"));
				}
			});
			this._worker.onerror = dojo.hitch(this, function(e){
				this._onError({ _error: true, responseText: e.message, workerError: e });
			});
			
			//console.debug(this._debugPrefix + "Starting worker");
			//console.dir(dojo.fromJson(data));

			// Send Dojo base dir to the worker, so it can load a minimal dojo:
			var dojoDir;
			if(this.dojoDir){
				dojoDir = this.dojoDir;
			} else {
				dojoDir = dojo.baseUrl;
				// should end with 'dojo/', remove it:
				dojoDir = dojoDir.substr(0, dojoDir.length - "dojo/".length);
				// get the (hopefully) absolute URL of the page, so we can compute an absolute path to dojo:
				var docUrl = document.location.href;
				var slash = docUrl.lastIndexOf("/");
				if(slash > 0)
					docUrl = docUrl.substr(0, slash+1);
				dojoDir = docUrl + dojoDir;
			}
			// the worker must use a different code to load Dojo 1.7, so warn it:
			if(dojo.version.major > 1 || (dojo.version.major == 1 && dojo.version.minor >= 7))
				this._worker.postMessage("_isDojo17");
			this._worker.postMessage(dojoDir); // load dojo

			// Send ibm_graphlayout module path:
			var glPath = dojo.config.modulePaths ? (dojo.config.modulePaths['ibm_ilog.graphlayout'] || '') : '';
			this._worker.postMessage(glPath);
		}
		
		// Now send the graph data:
		this._worker.postMessage(data);
	},
	
	_stop: function()
	{
		if(this._worker){
			this._worker.terminate();
			this._worker = null; // cannot be reused after terminate()
		}
	}
});
