/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2011, 2012. All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/
dojo.provide("rmpw.text.internal.util.Converter");

dojo.require("dojo.string");
dojo.require("dojox.string.Builder");

/**
 * Prototype implementation of an XHTML <--> JSON Conversion algorithm.
 * Currently used client-side as a part of Rich Text doc prototype.
 * Should be moved to server-side for performance and scalability reasons.
 */

rmpw.text.internal.util.Converter = {
	STRING_ATTRIBUTE_TEMPLATE: ' ${dmxml_key} = "${dmxml_stringValue}"',
	REF_ATTRIBUTE_TEMPLATE: ' ${dmxml_key} = "${dmxml_refValue.rdf_about}"',
	EXTERNAL_REF_ATTRIBUTE_TEMPLATE: ' ${dmxml_key} = "${dmxml_refValue.$ref}"',
	USE_RDF_ABOUT: false,
	
    parse: function(content){
        var xml = new dojox.string.Builder();
        if (this.isXhtmlRoot(content)) {
			this.serializeObject(content.dmxhtml_body, xml);
        } else {
			this.serializeObject(content, xml);
		}
        return xml.toString();
    },
    
    isXhtmlRoot: function(object){
        if (!object || !object.rdf_type) {
            return false;
        }
        return dojo.some(object.rdf_type, function(type) { return type.rdf_about === "http://jazz.net/ns/dm/document#Document"; });
    },
    
    isCDATA: function(object){
        if (!object || !object.rdf_type) {
            return false;
        }
        return dojo.some(object.rdf_type, function(type) { return type.rdf_about === "http://jazz.net/ns/dm/xml#CDATA"; });
    },
    
    serializeObject: function(object, xml){
        if (this.isCDATA(object)) {
            if (object.dmxml_text) {
                xml.append(object.dmxml_text);
            }
        }
        else {
            if (!object.dmxml_name) {
                return xml;
            }
            xml.append("<", object.dmxml_name);
            this.serializeAttributes(object, xml);

			if (this.USE_RDF_ABOUT && object.rdf_about) {
				xml.append(" id=",object.rdf_about.substring(object.rdf_about.lastIndexOf("#") + 1));
			}

            if (object.dmxml_content) {
                xml.append(">");
                var i, len = object.dmxml_content.length;
                for (i = 0; i < len; i++) {
                    this.serializeObject(object.dmxml_content[i], xml);
                }
                xml.append("</", object.dmxml_name, ">");
            } else if (object.dmxhtml_body) {
				this.serializeObject(object.dmxhtml_body, xml);
                xml.append("</", object.dmxml_name, ">");				
			}
            else {
                xml.append("/>");
            }
        }
		return xml;
    },
	
    serializeAttributes: function(object, xml){
        if (object.dmxml_attributes) {
            var i, attribute;
            for (i = 0; i < object.dmxml_attributes.length; i++) {
				attribute = object.dmxml_attributes[i];
				// Handle string attributes (attributes whose key/value pair are both strings)
				if (attribute.dmxml_key && attribute.dmxml_stringValue) {
					xml.append(dojo.string.substitute(this.STRING_ATTRIBUTE_TEMPLATE, attribute));
				}
				// Handle ref attributes (attributes whose value is an object with an rdf:about pointing to the referenced resource).
				// See the dmxml.owl ontology definition and the function comments on isRefAttribute() below.
				else if (attribute.dmxml_key && attribute.dmxml_refValue) {
					if (attribute.dmxml_refValue.rdf_about) {
						xml.append(dojo.string.substitute(this.REF_ATTRIBUTE_TEMPLATE, attribute));					
					} else {
						xml.append(dojo.string.substitute(this.EXTERNAL_REF_ATTRIBUTE_TEMPLATE, attribute));
					}
				}
            }
        }
		return xml;
    },


// XHTML to JSON


    parseToJson: function(xhtml, resourceUri){
        var stack = [];
        var root = this.createXhtmlRoot(null, resourceUri);
        var parser = new CKEDITOR.htmlParser();
		var context = this.createXhtmlBody(root, null);
		root.dmxhtml_body = context;

        parser.onTagOpen = dojo.hitch(this, function(tagName, attributes, selfClosing){
            var tag = null;
            if (tagName === "html") {
                tag = this.createXhtmlRoot(attributes, resourceUri);
				tag = dojo.mixin(root, tag);
            } else if (tagName === "body") {
				tag = this.createXhtmlBody(root, attributes);
				tag = dojo.mixin(context, tag);
				root.dmxhtml_body = tag;
			} else {
                context.dmxml_content = context.dmxml_content || [];
                tag = this.createTag(context, tagName, attributes);
                context.dmxml_content.push(tag);
                stack.push(context);
            }
            context = tag;
        });

        parser.onTagClose = dojo.hitch(this, function(tagName) {
            context = stack.pop();
        });

        parser.onText = dojo.hitch(this, function(text){
            context.dmxml_content = context.dmxml_content || [];
            var cdata = this.createCDATA(context, text);
            context.dmxml_content.push(cdata);
        });
		
        parser.parse(xhtml);
        return root.dmxhtml_body;
    },

    createXhtmlRoot: function(attributes, resourceUri){
        var root = dojo.mixin(this.createTag(null, "html", attributes), {
	        rdf_type: [
				{ rdf_about: "http://jazz.net/ns/dm/document#Document"},
				{ rdf_about: "http://jazz.net/ns/dm/core#Document"}
			],
			rdf_about: ""
        });
        return root;
    },

    createXhtmlBody: function(root, attributes){
        var body = dojo.mixin(this.createTag(root, "body", attributes), {
            rdf_type: [{ rdf_about: "http://jazz.net/ns/dm/document#Body" }]
        });
        return body;
    },

    createTag: function(context, tagName, attributes) {
        var tag = this.createRdfObject("http://jazz.net/ns/dm/xml#BodyTag");
		if (tagName) {
			tag.dmxml_name = tagName;
		}
		if (attributes) {
			if (this.USE_RDF_ABOUT && attributes.id) {
				tag.rdf_about = attributes.id;
			}
		}
		
		if (this.USE_RDF_ABOUT && !tag.rdf_about && context) {
			tag.rdf_about = context.rdf_about + "T" + String(context.dmxml_content.length);
		}

        for (var att in attributes) {
			if (!(this.USE_RDF_ABOUT && att !== "id")) {
				var attrib = this.createAttribute(tag, att, attributes[att]);
	            tag.dmxml_attributes = tag.dmxml_attributes || [];			
				tag.dmxml_attributes.push(attrib);
			}
        }
        return tag;
    },

    createAttribute: function(context, key, value) {
		var isRefAttribute = this.isRefAttribute(context, key);
        var attrib = this.createRdfObject(
			isRefAttribute ?
			"http://jazz.net/ns/dm/xml#RefAttribute" :
			"http://jazz.net/ns/dm/xml#StringAttribute");
		attrib.dmxml_key = key;
        if (isRefAttribute) {
			attrib.dmxml_refValue = {rdf_about: value};
		} else {
			attrib.dmxml_stringValue = value;
		}
		if (this.USE_RDF_ABOUT && context) {
			attrib.rdf_about = context.rdf_about + "A" + String (context.dmxml_attributes.length);			
		}
		return attrib;
    },

	/**
	 * Determines whether or not a given attribute is a reference to another resource.
	 * Some references in XHTML (eg: the href attribute of an <a> tag) are references
	 * to other resources.  These need to be handled differently on serialization
	 * and deserialization since the XML ontology represents these differently.
	 * 
	 * Currently we support only <img> and <a> tags, although there are many other XHTML
	 * tags that can have reference attributes.  @TODO We should visit these on a case-by-case
	 * basis to determine if they should be exposed as resource references to the system.
	 * 
	 * @param {Object} context
	 * @param {Object} key
	 */
	isRefAttribute: function(context, key) {
		var name = context.dmxml_name;
		if (name === 'img') {
			return key === 'src';
		} else if (name === 'a') {
			return key === 'href';
//		} else if (name === 'object') {
			//TODO need to define exactly how our embed objects are going to specify the URI for the embedded resource.
		}
		return false;
	},
	
    createCDATA: function(context, text){
        return dojo.mixin(this.createTag(context), {
			rdf_type: [{ rdf_about: "http://jazz.net/ns/dm/xml#CDATA" }],
            dmxml_text: text
        });
    },
    
    createRdfObject: function(rdf_types){
        if (dojo.isArray(rdf_types)) {
            rdf_types = dojo.map(rdf_types, function(type) {
				return {
					rdf_about : type
				};
			});
        } else {
            rdf_types = [{ rdf_about: rdf_types }];
        }
        return {
            rdf_type: rdf_types
        };
    },

    htmlifyText: function(source){
        // space out rdf:Description blocks
        source = source.replace(/<rdf:Description/g, "\n  <rdf:Description");
        
        // escape angle brackets
        source = source.replace(/</g, "&lt;");
        source = source.replace(/>/g, "&gt;");
        source = "<html xmlns='http://www.w3.org/1999/xhtml/'><pre><code>" + source + "</code></pre></html>";
        return source;
    }
};
