/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.policies.GraphEdge");

dojo.require("rmpw.diagrams.render.policies.GraphContainer");
dojo.require("rmpw.diagrams.render.policies.GraphElement");
dojo.require("rmpw.diagrams.render.graph.AnimatedEdge");
dojo.require("rmpw.diagrams.render.util.PointListUtilities");
dojo.require("rmpw.diagrams.render.util.Workarounds");
dojo.require("rmpw.diagrams.render.util.PointList");
dojo.require("rmps.core.error");

/**
 * Data needed to layout the graph edge
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.policies.GraphEdge", rmpw.diagrams.render.policies.GraphElement, {
	singleton: ["graphElement", "graphEdge"],
	
	sourceMapping: null,
	targetMapping: null,
	
	constructor: function(mapping) {
		// TODO (MG): find another solution to this, level is then no longer a true tree level, can't be counted
        //       on for algorithms that make use of level equality (like finding lowest common ancestor)
        this.host.level += 0.5;
        // push edge level to be greater then the default so they are processed after

		
		this._unpositioned = true;
	},
	
	_invalidate: function() {
		this.inherited(arguments);
		
		this.getInternalGraphElement().invalidate();
	},
	
	_completeNotationTraces: function() {
		this.inherited(arguments);
		
		// determine which aspects of notation aren't traced yet and add internal traces to complete
		var internalTraces = [];
		
		this._createInternalTrace("sourceAnchor", null, internalTraces);
		this._createInternalTrace("targetAnchor", null, internalTraces);
		
		// default value for router trace will mach that defined in QVT, or Manual if default not defined
		var graphSource = this.host.sources.__metadata;		
		if (!graphSource || !graphSource.router || graphSource.router === "oblique") {
			this._createInternalTrace("router", "Manual", internalTraces);
		} else {
			this._createInternalTrace("router", "Rectilinear", internalTraces);
		}
		
		this._createInternalTrace("bendpoints", [], internalTraces);
		
		if (internalTraces.length > 0) {
			this.host.connectTraces(this.host.sources.__metadata, internalTraces);
		}
	},
	
	destroy: function() {
	    var anchorable;
		if(this.sourceMapping) {
			anchorable = this.getAnchorable(this.sourceMapping);
			if (anchorable) {
				anchorable.removeSourceConnection(this.host);
			}
		}
		
		if(this.targetMapping) {
			anchorable = this.getAnchorable(this.targetMapping);
			if (anchorable) {
				anchorable.removeTargetConnection(this.host);
			}
		}
		
		this._detachListeners(this._sourceMappingInvalidationListeners);
		this._detachListeners(this._targetMappingInvalidationListeners);
		
		this.inherited(arguments);
	},

	/* 
	 * Returns the dojo shape associated with this graph policy element
	 */
	getDojoShape: function() {
		return this._hitzone ? this._hitzone : this._polyline;
	},
	
	setPolyline: function(polyline) {
		this._polyline = polyline;
		this._setVisibility(this._polyline, !this._unpositioned);
	},
	
	setHead: function(polyline) {
		if (dojo.isString(polyline)) {
			this._headTemplate = polyline;
		} else {
			this._head = polyline;
			this._setVisibility(this._head, !this._unpositioned);
		}
	},
	
	setTail: function(polyline) {
		if (dojo.isString(polyline)) {
			this._tailTemplate = polyline;
		} else {
			this._tail = polyline;
			this._setVisibility(this._tail, !this._unpositioned);
		}
	},
	
	setSmoothness: function(smoothness) {
		switch(smoothness) {
		case "none":
			this._smoothness = 0; 
			break;
		case "less":
			this._smoothness = 15; 
			break;
		case "normal":
			this._smoothness = 30; 
			break;
		case "more":
			this._smoothness = 45; 
			break;
		}
	},

	/*
	 * @deprecated - use setShape instead
	 */
	setFigure: function(figure) {
		this.setShape(figure);
	},
	
	setShape: function(figure) {
		this.inherited(arguments);
		
		if(dojox.gfx.svg) {
			// a thicker, hidden line for catching mouse hits to make the connection easier to select
			this._hitzone = figure.createPolyline();
			this._hitzone.setStroke({color: [255, 0, 0, 0], width: 10});
			/*
			 * in firefox 3.6, they implemented svg's pointer-events property:
			 * - http://www.w3.org/TR/SVG11/interact.html#PointerEventsProperty
			 * - https://bugzilla.mozilla.org/show_bug.cgi?id=563500 (bug about change from 3.5 > 3.6)
			 * 
			 * This requires us to set pointer-events explicitly to "stroke", or the zero opacity above
			 * will disqualify us from events under the default "visiblePainted" rules
			 */
		
			this._hitzone.rawNode.setAttribute("pointer-events", "stroke");
		}
		
		var mapping = this.host;
		var that = this;
		dojo.subscribe('gef/diagram/avoidObstructions', dojo.hitch(this, function(avoid) {
			that.getInternalGraphElement().setAvoidObstructions(avoid);
			mapping.invalidate();
		}));
	},
	
	_createInternalGraphElement : function() {
		return this.host.viewer.graphConfigurator.createGraphEdge(this.host, 
				this.getGraphContainer().getInternalGraphElement(), 
				this.sourceMapping ? this.sourceMapping.policies.graphElement.getInternalGraphElement() : null, 
				this.targetMapping ? this.targetMapping.policies.graphElement.getInternalGraphElement() : null);
	},

	_getAttachedGraphElement: function(viewMappings, isSource) {
		if (viewMappings) {
			for (var i=0, viewMapping; viewMapping = viewMappings[i]; i++) {
				var anchorable = this.getAnchorable(viewMapping);
				if (anchorable && (viewMapping.root === this.host.root) && 
					// if source make sure this mapping doesn't already have this edge
					!this._hasThisEdge(isSource ? anchorable.ownedConnections : anchorable.targetConnections) ) {
					return viewMapping;
				}
			}
		}
		
		return null;
	},

	_hasThisEdge: function(connections) {
		for (var i=0, edgeMapping; edgeMapping=connections[i]; i++) {
			if (edgeMapping.hasSameModelSource(this.host) && this.host.declaredClass.localeCompare(edgeMapping.declaredClass) === 0) {
				return true;
			}
		}
		return false;
	},
	
	/*
	 * method that gets called for an extra logic required by graph elements for invalidation
	 */
	enforce: function() {
		var viewMapping;
		if (this._sourcesTarget && this._sourcesTarget.__mappings) {
			viewMapping = this._getAttachedGraphElement(this._sourcesTarget.__mappings);
			if(viewMapping) {
				this._setTargetMapping(viewMapping);
				delete this._sourcesTarget;
			} else {
				this.host.invalidate();
				return;
			}
		}
		
		if (this._sourcesSource && this._sourcesSource.__mappings) {
			viewMapping = this._getAttachedGraphElement(this._sourcesSource.__mappings, true);
			if(viewMapping) {
				this._setSourceMapping(viewMapping);
				delete this._sourcesSource;
			} else {
				this.host.invalidate();
				return;
			}
		}

		// move this edge's dojo shape to diagram if it already isn't there
		// needed to make translation methods work

        if(!this._isInitialized && this.sourceMapping && this.targetMapping && this._shape) {
            this._isInitialized = true;
            var edgeParentShape = this._shape.parent;
            var diagramShape = this.host.viewer.diagram.getShape();
            if(edgeParentShape !== diagramShape) {
                this._isInDiagram = true;
                if (dojo.isIE != 8) {
                    edgeParentShape.remove(this._shape);
                    diagramShape.add(this._shape);
                } else {
                	// doing this immediately with IE8 causes the browser to terminate :(
	        		var self = this;
	        		this._callbackTimeout = setTimeout( function() {
	        			if (self._callbackTimeout) {
	        				clearTimeout(self._callbackTimeout);
	        			}
	                    edgeParentShape.remove(self._shape);
	                    diagramShape.add(self._shape);
	                    self.host.invalidate();
	             		}, 0);
                }
            }
        }	

        this._refreshDecoration("_head", "_headTemplate");
        this._refreshDecoration("_tail", "_tailTemplate");
		
		this.inherited(arguments);
		
		// Note source and target must be set before source anchor and target anchor can be set
		this._enforceInternalTrace("sourceAnchor", "setSourceAnchor");
		this._enforceInternalTrace("targetAnchor", "setTargetAnchor");
		this._enforceInternalTrace("router", "setRouterType");
		this._enforceInternalTrace("bendpoints", "setBendpoints");
	},
	
	_refreshDecoration: function(shape, temp) {
		if (this._polyline) {
			var rawTemplate = this[temp];
			if (rawTemplate) {
				// if edge already has a decoration, remove it
				if (this[shape]) {
					this[shape].removeShape();
				}
				// use template object to create new decoration 
				var template = dojo.fromJson(rawTemplate); 
				if (template.path || template.circle || template.rectangle) {
					var figure;
					if (template.path) {
						figure = this._polyline.getParent().createPath({path: template.path});
					} else if (template.circle) {
						figure = this._polyline.getParent().createCircle({cx: template.circle.cx, cy: template.circle.cy, r: template.circle.r});
					} else {
						figure = this._polyline.getParent().createRect({x: template.rectangle.x, y: template.rectangle.y, width: template.rectangle.width, height: template.rectangle.height });
					}
					if (figure) {
						if (template.strokeColor) figure.setStroke({color: template.strokeColor, width: template.strokeWidth});
						if (template.fillColor) figure.setFill(template.fillColor);
						this[shape] = figure;
					}
				} else {
					this[shape] = null;
				}
				delete this[temp];
			}
		}
	},
	
	
	isInDiagram: function() {
        if (!this._isInDiagram && this.sourceMapping && this.targetMapping && this._shape) {
            var edgeParentShape = this._shape.parent;
            var targetParentShape = this.targetMapping.getShape().parent;
            if(edgeParentShape !== targetParentShape) {
                var diagramShape = this.host.viewer.diagram.getShape();
                if(edgeParentShape !== diagramShape) {
                    this._isInDiagram = true;
                }
            }
        }

        return this._isInDiagram;
	},
	
	getState: function() {
		var state = {};
		
		state.sourceAnchor = this._getInternalTraceData("sourceAnchor");
		state.targetAnchor = this._getInternalTraceData("targetAnchor");
		state.router = this._getInternalTraceData("router");
		state.bendpoints = dojo.clone(this._getInternalTraceData("bendpoints"));
		return state;
	},
	
	setState: function(state) {
		this.setSourceAnchor(state.sourceAnchor);
		this.setTargetAnchor(state.targetAnchor);
		this.setRouterType(state.router);
		this.setBendpoints(state.bendpoints);
		
		this.getInternalGraphElement().invalidate();
	},
	
	getAnimatedFigure: function() {
		if (this.isInitialized()) {
			return new rmpw.diagrams.render.graph.AnimatedEdge(this);
		}
		
		return null;
	},

	getRouterType: function() {
		return this.getInternalGraphElement().getRouterType();
	},
	
	setRouterType: function(router) {
		if (router !== undefined) {
			this.host.sources.__metadata.router = router;
		}
		
		this.getInternalGraphElement().setRouterType(router);
	},
	
	setBendpoints: function(bendpoints) {
		if (bendpoints !== undefined) {
			// nothing in rdf yet
			if (bendpoints === null) {
				bendpoints = [];
			} else if (dojo.isString(bendpoints)) {
				//convert back into array
				var bendpts = bendpoints.split(" ");
				bendpoints = [];
				for (var j = 0, lg = bendpts.length; j < lg; ) {
					bendpoints.push({sourceX: parseFloat(bendpts[j++]), sourceY: parseFloat(bendpts[j++]), 
						targetX: parseFloat(bendpts[j++]), targetY: parseFloat(bendpts[j++])});
				}
			}
			this.host.sources.__metadata.bendpoints = bendpoints;
			if (this.host.sources.__metadata.bendpoints.length>0) {
				this.getInternalGraphElement().setAvoidObstructions(false);
			}
		}
		
		this.getInternalGraphElement().setRouterConstraint(bendpoints, true);
	},
	
	setAvoidObstructions: function(avoidObstructions) {
		if (avoidObstructions !== undefined) {
        	if (dojo.isString(avoidObstructions)) avoidObstructions = avoidObstructions == "true";
    		this.getInternalGraphElement().setAvoidObstructions(avoidObstructions);

        	// for performance, lock in bendpoints
        	this._avoidObstructions = avoidObstructions;
    		this.refreshAvoidance();
		}
	},
	
	refreshAvoidance: function() {
		// PolylineBenpointEditable listens to this
	},
	
	setMaxBendpoints: function(maxBendpoints) {
		this.maxBendpoints = maxBendpoints;
	},
	
	getMaxBendpoints: function() {
		return this.maxBendpoints;
	},
	
	setSourceAnchor: function(sourceAnchor) {
		if (sourceAnchor !== undefined) {
			this.host.sources.__metadata.sourceAnchor = sourceAnchor;
		}
		
		var graphEl = this.getInternalGraphElement();
		if (graphEl.getSourceAnchor()) {
			if (sourceAnchor === null) {
				graphEl.getSourceAnchor().setRelativeReferencePoint(null);
			}
			else {
				graphEl.getSourceAnchor().setRelativeReferencePoint(graphEl.getSourceAnchor().parseTerminalString(sourceAnchor));
			}
		}
	},
	
	setTargetAnchor: function(targetAnchor) {
		if (targetAnchor !== undefined) {
			this.host.sources.__metadata.targetAnchor = targetAnchor;
		}
		
		var graphEl = this.getInternalGraphElement();
		if (graphEl.getTargetAnchor()) {
			if (targetAnchor === null) {
				graphEl.getTargetAnchor().setRelativeReferencePoint(null);
			}
			else {
				graphEl.getTargetAnchor().setRelativeReferencePoint(graphEl.getTargetAnchor().parseTerminalString(targetAnchor));
			}
		}
	},
	
	_setSourceMapping: function(sourceMapping) {
		this._detachListeners(this._sourceMappingInvalidationListeners);
				
		this.sourceMapping = sourceMapping;
		this.getInternalGraphElement().setSource(this.sourceMapping.policies.graphElement.getInternalGraphElement());
		var anchorable = this.getAnchorable(sourceMapping);
		if (anchorable) {
			this.getInternalGraphElement().setSourceAnchor(anchorable.getSourceConnectionAnchor(this.host));
			anchorable.addSourceConnection(this.host);
			this._addListeners(this._sourceMappingInvalidationListeners, this.sourceMapping);
		}
		else {
			console.debug("source isn't anchorable");
		}
	},
	
	setSource: function(source) {
		var realSource = this.host.getResolvedSemantic(source);
		
		if(dojo.isArray(realSource)){
		    console.error("Array of sources not supported by GraphEdge, attempting to recover by using first element.");
			realSource = source[0];
		}
		if(realSource) {
			if (this.sourceMapping ) {
				var anchorable = this.getAnchorable(this.sourceMapping);
				if (anchorable) {
					anchorable.removeSourceConnection(this.host);
				}
			}
			
			if (realSource) {
				var viewMapping = this._getAttachedGraphElement(realSource.__mappings, true);
				if (viewMapping) {
					this._setSourceMapping(viewMapping);
				}
				else {
					this._sourcesSource = realSource;
				}
			}
		}
	},
	
	_setTargetMapping: function(targetMapping) {
		this._detachListeners(this._targetMappingInvalidationListeners);
		
		this.targetMapping = targetMapping;
		this.getInternalGraphElement().setTarget(this.targetMapping.policies.graphElement.getInternalGraphElement());
		var anchorable = this.getAnchorable(targetMapping);
		if (anchorable) {
			this.getInternalGraphElement().setTargetAnchor(anchorable.getTargetConnectionAnchor(this.host));
			anchorable.addTargetConnection(this.host);
			this._addListeners(this._targetMappingInvalidationListeners, this.targetMapping);
		}
		else {
			console.debug("target isn't anchorable");
		}
	},
	
	setTarget: function(target) {
		var realTarget = this.host.getResolvedSemantic(target);
		
		if(dojo.isArray(realTarget)){
			console.error("Array of realTargets not supported by GraphEdge, attempting to recover by using first element.");
			realTarget = target[0];
		}
		
		if (this.targetMapping ) {
			var anchorable = this.getAnchorable(this.targetMapping);
			if (anchorable) {
				anchorable.removeTargetConnection(this.host);
			}
		}
		
		if (realTarget) {
			var viewMapping = this._getAttachedGraphElement(realTarget.__mappings);
			if (viewMapping) {
				this._setTargetMapping(viewMapping);
			}
			else {
				this._sourcesTarget = realTarget;
			}
		}
	},

	// listen to invalidation on mapping and any non-diagram container mappings
	_addListeners: function(listeners, mapping) {
		listeners = [];
		listeners.push(dojo.connect(mapping, "invalidate", this.host, "invalidate"));
		this._addParentListeners(listeners, mapping.parent);
	},
	
	// if a parent of this source/target is invalidated, we wait for it to be refreshed before invalidating
	// edge so that all parent nodes have their positions before the edge is routed
	_addParentListeners: function(listeners, mapping) {
		if (mapping && mapping.policies.graphContainer && !mapping.policies.graphContainer.isTopLevel()) {
			listeners.push(dojo.connect(mapping, "invalidate", dojo.hitch(this, function() {
				this.host.invalidate();					
			})));
			this._addParentListeners(listeners, mapping.parent);
		}
	},
	
	_detachListeners: function(listeners) {
		if (listeners) {
			dojo.forEach(listeners, dojo.disconnect);
			delete listeners;
		}
	},
	
	/*
	 * Gets the source connection anchor for this edge
	 */
	getSourceConnectionAnchor: function() {
		return this.getInternalGraphElement().getSourceAnchor();
	},
	
	/*
	 * Gets the target connection anchor for this edge
	 */
	getTargetConnectionAnchor: function() {
		return this.getInternalGraphElement().getTargetAnchor();
	},
	
	/*
	 * Returns boolean is true if this edge is orthogonal (straight) or false otherwise
	 */
	isOrthogonal: function() {
		return this.getInternalGraphElement().isOrthogonal();
	},
	
	/*
	 * Retrieve the polygon points - used for outline view and moveable anchor
	 */
	getPolygonShapes: function() {
		return [this.getPointList()];
	},
	
	/*
	 * Gets the points associated with this edge
	 */
	getPointList: function(smooth) {
		var pointList = this.getInternalGraphElement().getPointList();
		if ((smooth === undefined || smooth) && this._smoothness) {
			var util = new rmpw.diagrams.render.util.PointListUtilities();
			pointList = util.calcSmoothPolyline(pointList, this._smoothness, 15);
		}
		return pointList;
	},
	
	setPointList: function (points){
		var oldPoints = this.getDojoPoints();
		if (!this._oldPoints) {
			this._oldPoints = oldPoints;
		}
		this.updateShape();		
		this.getGraphManager().updateLocation(this.host);
	},
	
	_calculateAngle: function(p1x, p1y, p2x, p2y) {
		var dx = Math.abs(p2x - p1x);
		var dy = p2y - p1y;
		if(dx === 0 && dy === 0){ return 0; }
		var angle = Math.atan(dy/dx)/Math.PI*180;
		if(p2x - p1x < 0){
			angle = 180 - angle;
		}
		return angle;
	},
	
	/*
	 * Returns the bounds of the dojo shape
	 */
	getDojoPoints: function() {
		if (this._oldPoints) {
			return this._oldPoints;
		}
		
		var pts = this._polyline.shape.points;
		var relPts = new rmpw.diagrams.render.util.PointList();
		for (var i=0, pt; pt=pts[i]; i++) {
			relPts.addPoint_0(pt.x, pt.y);
		}
		return relPts;
	},
	
	updateShape: function () {
		try{
			var graphEl = this.getInternalGraphElement();
			if ((!this._polyline && !this._hitzone) || !graphEl.getTargetAnchor() || !graphEl.getSourceAnchor()) {
				return;
			}
			
			if (!this.isUsingDojoPosition()) {
				var pointList = this.getPointList();
				if (!this._smoothness && this.isOrthogonal()) {
					// orthogonal edge will have rounded corners (not customizable for now)
                    var util = new rmpw.diagrams.render.util.PointListUtilities();
					var rForBendpoint = [];
					var pl = util.calcRoundedCornersPolyline(pointList, 5, rForBendpoint, true);
					if (pl) {
						// just in case something goes wrong use the default polyline
						pointList = pl;
					}
				}
				var points = pointList.points;
				this._polyline.setShape(points);
				if (this._hitzone) {
					this._hitzone.setShape(points);
				}
		
				if(this._tail) {
					var firstPoint = points[0];
					var secondPoint = points[1];
					this._tail.setTransform({dx: firstPoint.x, dy: firstPoint.y});
					this._tail.applyTransform(dojox.gfx.matrix.rotategAt(
							this._calculateAngle(secondPoint.x, secondPoint.y, firstPoint.x, firstPoint.y), 0, 0));
				}
		
				if(this._head) {
					var len = points.length;
					if (len > 1) {
						var lastPoint = points[len - 1];
						var secondLastPoint = points[len - 2];
						this._head.setTransform({dx: lastPoint.x, dy: lastPoint.y});
						this._head.applyTransform(dojox.gfx.matrix.rotategAt(
								this._calculateAngle(secondLastPoint.x, secondLastPoint.y, lastPoint.x, lastPoint.y), 0, 0));
					}
				}
			}
		}
		finally {
			if (this._oldPoints) {
				delete this._oldPoints;
			}
		}
	},
	
	refresh: function() {
		this.inherited(arguments);
		
		// set point list just updates this._oldPoints
		this.setPointList(this.getPointList());

		if (this._unpositioned && this.sourceMapping && this.targetMapping) {
			delete this._unpositioned;
			this._setVisibility(this._polyline, true);
			this._setVisibility(this._head, true);
			this._setVisibility(this._tail, true);
		}
	},
	
	_setVisibility: function(shape, visible) {
		if (shape  && !dojo.isOffscreen) {
			if (visible) {
				shape.rawNode.style.visibility = '';
			} else {
				shape.rawNode.style.visibility = dojo.isIE === 8 ? "collapse"  : "hidden";
			}
		}
	}

});