/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.policies.GraphNodeLabelLayout");

dojo.require("rmpw.common.policy.Policy");
dojo.require("rmpw.diagrams.render.util.Workarounds");
dojo.require('dojox.gfx.shape');

/**
 * Policy that positions and sizes labels within, above or below a sibling shape.
 * 
 * @author John Swanke
 */
dojo.declare("rmpw.diagrams.render.policies.GraphNodeLabelLayout", rmpw.common.policy.Policy, {
	singleton: ["resizeable", "graphNodeLabelLayout"],

	setLabel: function() {
		var label = arguments;
		if(label.length > 1) {
			console.warn("multiple labels not supported");
			return;
		}	
		var type = label[0].shape.type;
		if(type != "wrappingText") {
			console.warn("only wrappingText label supported");
			return;
		}
		this._label = label[0];
	},

	setShape: function() {
		var shape = arguments;
		if(shape.length > 1) {
			console.warn("multiple shapes not supported");
			return;
		}
		var type = shape[0].shape.type;
		if(type != "rect" && type != "image" && type!= "scalingImage") {
			console.warn("only rectangle, image and scaling image shapes supported");
			return;
		}
		this._shape = shape[0];
	},

	setBackground: function() {
		this._background = arguments[0];
		this._isFloatingLabel = this.host.policies.graphFloatingLabel !== undefined;
		
		// background is invisible but draggable
		if (!this._background.fillStyle) {
			this._background.setFill([255, 255, 255, 0.0]);
		}
	},

	// future
	setIcon: function() {
		this._icon = arguments;
	},
	
	// means shape is proportional and min size is used on longest side if
	//  shape has not yet been resized by user
	setProportionalShapeMinSize: function(proportionalShapeMinSize) {
		this._proportionalShapeMinSize = proportionalShapeMinSize;
	},
	
	setWidth: function(width) {
		this._width = width;
	},
	
	setHeight: function(height) {
		this._height = height;
	},
	
	setMinShapeWidth: function(minShapeWidth) {
		this._minShapeWidth = minShapeWidth;
	},
	
	setMinShapeHeight: function(minShapeHeight) {
		this._minShapeHeight = minShapeHeight;
	},	
	
	setLabelHPosition: function(labelHPosition) {
		this._labelHPosition = labelHPosition;
	},	
	
	setLabelVPosition: function(labelVPosition) {
		this._labelVPosition = labelVPosition;
	},	
	
	resize: function(box) {
		var cache = this._background.__labelLayoutCache; 
		if (cache) {
			if (this.host.policies.graphContainer) {
				var containerBounds = {x:0, y:0, width: cache.cwidth, height: cache.cheight};
			}
			// resize based on new string
			if (box === undefined) {
				this._layout({ x: 0, y: 0, width: this._width, height: this._height }, containerBounds, this._isUserSized);			
			} else {
			// resize based on new box
				var dx = box.dx * 1.0;
				var dy = box.dy * 1.0;
				var w = box.width*1.0;
				var h = box.height*1.0;
				if (w < 0.0) {
					w = -w;
					dx -= w;
				}
				if (h < 0.0) {
					h = -h;
					dy -= h;
				}
				
				// adjust for zoom
			    var matrix = this.host.root.viewport.contents.matrix;
			    dx/=matrix.xx;
			    dy/=matrix.xx;
			    h/=matrix.xx;
			    w/=matrix.xx;
				
				this._layout({ x: dx, y: dy, width: w, height: h }, containerBounds, true);			
			}
		}
	},
	
	enforce: function() {
		// at least one background
		if(!this._background) {
			console.warn("no background specified");
			return;
		}
		
		// don't enforce if nothing has changed
		var cache = this._background.__labelLayoutCache; 
		var labelText = this._label ? this._label.getText() :"";
		var imageSrc = this._shape ? (this._shape.getSrc ? this._shape.getSrc() : this._shape.src) : null;
		var hostWidth = this._width;
		var hostHeight = this._height;
		// if this shape has a container, adjust sizes accordingly
		if (this.host.policies.graphContainer) {
			var containerBounds = this.host.policies.graphContainer.getChildrenBox();
			var containerWidth = containerBounds.width;
			var containerHeight = containerBounds.height;
		}
		if(cache && (cache.text == labelText) && (cache.imageSrc == imageSrc) && 
				(cache.width == hostWidth) && (cache.height == hostHeight) && (cache.cwidth == containerWidth) && (cache.cheight == containerHeight) &&
				(cache.labelHPosition == this._labelHPosition) && (cache.labelVPosition == this._labelVPosition) &&
				(cache.defaultShapeHeight == this._defaultShapeHeight) && (cache.defaultShapeWidth == this._defaultShapeWidth) &&
				(cache.proportionalShapeMinSize == this._proportionalShapeMinSize)) {
			return;
		}
		this._background.__labelLayoutCache = {
				text: labelText,
				imageSrc: imageSrc,
				width: hostWidth,
				height: hostHeight,
				cwidth: containerWidth,
				cheight: containerHeight,
				labelVPosition: this._labelVPosition,
				labelHPosition: this._labelHPosition,
				minShapeHeight: this._minShapeHeight,
				minShapeWidth: this._minShapeWidth,
				proportionalShapeMinSize: this._proportionalShapeMinSize
				
			};
		
		this._isUserSized = (hostWidth!=null && hostHeight!=null );
		// in case property dialog converted to string
		if (dojo.isString(this._background.shape.width)) {
			this._background.shape.width = parseInt(this._background.shape.width);
		}
		if (dojo.isString(this._background.shape.height)) {
			this._background.shape.height = parseInt(this._background.shape.height);
		}
		var backgroundBounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(this._background, null, this._background.parent);
		// make sure background matches model
		if (hostWidth && hostWidth!=backgroundBounds.width) {
			this._background.setShape({width: hostWidth});
			backgroundBounds.width = hostWidth;
		}
		if (hostHeight && hostHeight!=backgroundBounds.height) {
			this._background.setShape({height: hostHeight});
			backgroundBounds.height = hostHeight;
		}
		this._layout(backgroundBounds, containerBounds, this._isUserSized);
	},
	
	_layout: function(backgroundBounds, containerBounds, isUserSized) {
		var theLabel = this._label; 
		var theShape = this._shape ? this._shape : null;
		var theBackground = this._background;

		// get padding
		if (theLabel && !this._isFloatingLabel) {
			var style = theLabel.customStyle;
			var padding = {
					left: (style && style.paddingLeft) ? style.paddingLeft : 8,
					top: (style && style.paddingTop) ? style.paddingTop : 2,
					right: (style && style.paddingRight) ? style.paddingRight : 8,
					bottom: (style && style.paddingBottom) ? style.paddingBottom : 8
				};
		} else {
			var padding = {left: 0,top: 0,right: 0,bottom: 0};
		}

		// proportional shape?
		var proportionalShapeMinSize = theShape  && theShape.getProportionalShapeMinSize ? theShape.getProportionalShapeMinSize() : null;
		if (!proportionalShapeMinSize) proportionalShapeMinSize = this._proportionalShapeMinSize;
		if (proportionalShapeMinSize) {
			var defaultImage = theShape.getDefaultImage ? theShape.getDefaultImage() : 	theShape.shape;
			// retry until image is loaded
			var w = defaultImage.width;
			var h = defaultImage.height;
			if (w == h && h == 0) {
				if (!this._timer) {
					if (this._timerCnt) { this._timerCnt++; } else { this._timerCnt=1; }
					// never going to load?
					if (this._timerCnt < 10) {
						this._timer = setTimeout(dojo.hitch(this, function() {
							delete this._timer;
							this.host.invalidate();
							this._background.__labelLayoutCache = null;
						}), 25);
					}
				}
				w = h = 1;
			}
			var defaultImageSize = {width:w, height:h};
		}

		// if special middleToBelow position, if custom image is defined, move label to below
		var isCustomImage = (theShape && theShape.isCustomImage && theShape.isCustomImage());
		var containerWithContent = containerBounds && containerBounds.width>1;
		var labelVPosition = this._labelVPosition;
		if (containerWithContent && (labelVPosition == 'middle' || labelVPosition == 'center' ||  labelVPosition == 'middleToBelow')) {
			labelVPosition = 'top';
		} else if (labelVPosition == 'middleToBelow') {
			labelVPosition = isCustomImage ? 'below' : 'middle';
		}
		// force label to top if container isn't empty
		// if proportional shape (ex: circle) make sure container size is applied proportionally too
		if (proportionalShapeMinSize && !isCustomImage && containerWithContent) {
			if (containerBounds.width > containerBounds.height) {
				containerBounds.height = containerBounds.width; 
			} else {
				containerBounds.width = containerBounds.height; 
			}
		}	
		var isOutside = labelVPosition == 'above' || labelVPosition == 'below';
		
		// enforce a minimum size user can make background
		// can't be less then container size or 32x32 whichever is bigger
		var labelBounds = theLabel ? theLabel.getUnwrappedBoundingBox() : {width:0, height:0};
		if (isUserSized && theLabel) {
			if (containerBounds && backgroundBounds.width < containerBounds.width + 10) {
				backgroundBounds.width = containerBounds.width + 10;
				theBackground.setShape({width: backgroundBounds.width});				
			} else if (backgroundBounds.width < 32) {
				backgroundBounds.width = 32;
				theBackground.setShape({width: backgroundBounds.width});
			}
			if (containerBounds && backgroundBounds.height < containerBounds.height + 10) {
				backgroundBounds.height = containerBounds.height + 10;
				theBackground.setShape({height: backgroundBounds.height});				
			} else {
				var bot = labelBounds.h + (isOutside? 32 :4);
				if (backgroundBounds.height < bot) {
					backgroundBounds.height = bot;
					theBackground.setShape({height: backgroundBounds.height});
				}
			}
		}
		var shapeBounds = dojo.clone(backgroundBounds);
		
		// set initial sizes to minimum
		if (!isUserSized) {
			var w,h;
			// if shape is proportional, get initial width/height using proportionalShapeMinSize && proportion
			if (proportionalShapeMinSize && defaultImageSize) {
				if (containerWithContent) {
					w = containerBounds.width + 20;
					h = containerBounds.height + 20;
				} else if (defaultImageSize.width > defaultImageSize.height) {
					w = proportionalShapeMinSize;
					h = w * defaultImageSize.height / defaultImageSize.width;
				} else {
					h = proportionalShapeMinSize;
					w = h * defaultImageSize.width / defaultImageSize.height;
				}
			} else if (containerWithContent && isOutside) {
				w = containerBounds.width + 20;
				h = containerBounds.height + 20;
			} else {
				w = (theShape ? this._minShapeWidth : (labelBounds.w + padding.left + padding.right));
				h = (theShape ? this._minShapeHeight : (labelBounds.h + padding.top + padding.bottom));
			}
			backgroundBounds.width = shapeBounds.width = w; 
			backgroundBounds.height = shapeBounds.height = h;
			if (isOutside && !containerWithContent) {
				backgroundBounds.height += labelBounds.h + 2;
			} else if (containerBounds) {
				// make as big as container and center shapes
				var topLeftPt = this.host.policies.graphContainer.getTopLeftPoint();
				var rightMargin = this._scale(topLeftPt.x,100);
				var bottomMargin = this._scale(topLeftPt.y,50);
				if (backgroundBounds.width < containerBounds.width + 10) backgroundBounds.width = shapeBounds.width = containerBounds.width + rightMargin;
				if (backgroundBounds.height < containerBounds.height + 10) backgroundBounds.height = shapeBounds.height = containerBounds.height + bottomMargin;

			}
			theBackground.setShape({width: backgroundBounds.width, height: backgroundBounds.height});
		} else if (containerWithContent && isOutside) {
			w = containerBounds.width + 20;
			h = containerBounds.height + 20;
		} else if (proportionalShapeMinSize && !containerWithContent) {
			shapeBounds = this._getProportionalShapeBounds(defaultImageSize, shapeBounds, backgroundBounds, labelBounds.h);
		}
		var clientBounds = dojo.clone(backgroundBounds);
		if (!isOutside) {
			clientBounds.x += padding.left;
			clientBounds.width -= padding.left;
			clientBounds.width -= padding.right;
			clientBounds.y += padding.top;
			clientBounds.height -= padding.top;
			clientBounds.height -= padding.bottom;
		}
		
		// do the layout
		
		if (theLabel) {
			// label controls width of shape 
			var x;
			if (!isUserSized && (labelBounds.w > clientBounds.width || !theShape) ){
				// if label has new lines, we need to first figure out the size of the label
				var wrappedLabelBounds = dojo.clone(clientBounds);
				wrappedLabelBounds.width = wrappedLabelBounds.height = -1;
				theLabel.applyLayout(wrappedLabelBounds, this._labelHPosition);
				var labelWidth = theLabel.getBoundingBox().width;
				if (isOutside) {
					x = 0;
					shapeBounds.x = (labelWidth - shapeBounds.width)/2;
					backgroundBounds.width = labelWidth;
				} else {
					x = padding.left;
					shapeBounds.width = labelWidth + 2*x;
					backgroundBounds.width = shapeBounds.width;
				}
				theBackground.setShape({width: backgroundBounds.width});
			} else {
			// else shape controls width of label-- or label centers text itself
				x = padding.left;
				var wrappedLabelBounds = dojo.clone(clientBounds);
				if (isOutside) {
					x = 0;
					wrappedLabelBounds.height -=32;
				}
				theLabel.applyLayout(wrappedLabelBounds, this._labelHPosition);
			} 
		
		// position label within shape
			var y;
			var labelHeight = theLabel.getBoundingBox().height;
			if (isOutside) {
				if (!isUserSized) {
					backgroundBounds.height = shapeBounds.height + labelHeight + 2;
					theBackground.setShape({height: backgroundBounds.height});
				} else {
					if (defaultImageSize && !containerWithContent) {
						shapeBounds = this._getProportionalShapeBounds(defaultImageSize, shapeBounds, backgroundBounds, labelHeight);
					} else {
						shapeBounds.height -= labelHeight;
					}
				}
				switch(labelVPosition) {
				case "above":
					y = 0;
					shapeBounds.y = labelHeight + 2;
					break;
				case "below":
					y = shapeBounds.y + shapeBounds.height - backgroundBounds.y + 2;
					break;
				}
			} else if (labelHeight <= clientBounds.height) {
				switch(labelVPosition) {
				case "above":
					y = (clientBounds.height - shapeBounds.height - labelHeight)/2;
					shapeBounds.y = labelHeight + 2;
					break;
				case "below":
					y = (clientBounds.height - shapeBounds.height - labelHeight)/2 + shapeBounds.height;
					break;
				case "start":
				case "top":
					y = padding.top; 
					break;
				case "end":
				case "bottom":
					y = backgroundBounds.height - padding.bottom - labelHeight;
					break;
				default:
				case "center":
				case "middle":
					y = (clientBounds.height - labelHeight) / 2 + padding.top;
					break;
				}
			} else {
				// position label at top of shape and increase size of shape
				y = padding.top;
				backgroundBounds.height = shapeBounds.height = labelHeight + padding.top + padding.bottom;
				theBackground.setShape({height: backgroundBounds.height});
			}
			
			// position label on background
			theLabel.setTransform({dx: x + backgroundBounds.x, dy: y + labelBounds.h + backgroundBounds.y - 2});
		}
		
		// if shape is always proportional (circle), change shape transform and background width
		if (!isOutside && proportionalShapeMinSize && !containerWithContent) {
			shapeBounds = this._getProportionalShapeBounds(defaultImageSize, shapeBounds, backgroundBounds, 0);
		}
		
		// position shape on background
		if (theShape) {
			theShape.setTransform({dx: shapeBounds.x, dy: shapeBounds.y});			
			theShape.setShape({width: shapeBounds.width, height: shapeBounds.height, container:containerWithContent});
		}
	},
	
	_getProportionalShapeBounds: function(defaultImageSize, shapeBounds, backgroundBounds,  labelHeight) {
		shapeBounds.height = backgroundBounds.height - labelHeight - 2;
		shapeBounds.width = shapeBounds.height * defaultImageSize.width / defaultImageSize.height;
		if (shapeBounds.width < backgroundBounds.width) {
			shapeBounds.x = (backgroundBounds.width - shapeBounds.width)/2;
			if (shapeBounds.x < 0) shapeBounds.x = 0;
			shapeBounds.y = 0;
		} else {
			shapeBounds.width = backgroundBounds.width;
			shapeBounds.height = shapeBounds.width * defaultImageSize.height / defaultImageSize.width;
			shapeBounds.y = (backgroundBounds.height - shapeBounds.height - labelHeight)/2;
			if (shapeBounds.y < 0) shapeBounds.y = 0;
			shapeBounds.x = 0;
		}
		shapeBounds.x += backgroundBounds.x;
		shapeBounds.y += backgroundBounds.y;
		return shapeBounds;
	},
	
	_scale: function(margin, max) {
		if (margin < max ) {
			return margin > 10 ? margin : 10;
		} else if (margin > (max*2 - 10)) {
			return 10;
		} else {
			return max - (margin - max);
		}
	}
	
	
});