/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmpw.common.store.JfsResource");

/**
 * Used to track the creation of new JFS resources (documents mostly) 
 */
dojo.declare("rmpw.common.store.JfsResource", null, {
	
	resources: null,
	jfsStore: null,
	
	resourcesUpdates: null,
	
	constructor: function(jfsStore) {
		this.jfsStore = jfsStore;
		this.resources = [];
		this.resourcesUpdates = [];
	},
	
	addResource: function(model, types, namespaces, identifier, target, onResourceCreation){
		this.resources.push(new rmpw.common.store.JfsResource._Resource(model, types, namespaces, identifier, target, onResourceCreation));
	},
	
	getResource: function(identifier){
		for(var index = 0, resource; resource = this.resources[index]; index++) {
			if(resource.identifier === identifier){
				return resource;				
			}
		}
	},
	
	removeResource: function(identifier){
		for(var index = 0, resource; resource = this.resources[index]; index++) {
			if(resource.identifier === identifier|| resource=== identifier){
				this.resources.splice(index, 1);
				index = 0;	//re-loop after splicing an item in the array				
			}
		}
	},
	
	removeAll: function(){
		this.resources = [];
	},
	
	/**
	 * Called when upon Save being clicked on the stand-alone form. 
	 * Ensures that all resources created from the pop-up forms are saved.
	 */
	saveResources: function() {
		var deferreds = [];
		for(var i = 0, resource; resource = this.resources[i]; i++) {
			var saveCallback = this.jfsStore.saveResource(resource.model, {rdf_type: resource.types}, dojo.hitch(this, function(_resource, response, ioArgs){
				this.removeResource(_resource);
				// ensure that the reference to newly created resource is set
				_resource.onResourceCreation(ioArgs.xhr.getResponseHeader("logicalURI"), _resource);
			}, resource));
			deferreds.push(saveCallback);
		}
		return new dojo.DeferredList(deferreds);
	},
	
	/**
	 * To be called after the reference to the newly created resource is set on the pop-up form.
	 * These will be saved later (see saveAllResourceUpdates) 
	 */
	addResourceUpdate: function(changeSets) {		
		this.resourcesUpdates.push(changeSets);
	},
	
	/**
	 * To be called after all new resources from this JfsResource are created, so now we can save the updates to these resources 
	 * (update consists of setting a reference to another newly created resource).
	 */
	saveAllResourceUpdates: function() {
		// This array is used to track change sets based on the uri of the document being updated by them.
		// Server side expects each put request to be only for one document, so that is why we need to submit
		// a separate update request for each document (there could be more changes in one document though, in case
		// when more than one reference to a new resource is set on the document). 
		var resourceUrisWithUpdates = [];
		var logicalUri;
		var found;
		for (var i = 0, resUpdates; resUpdates = this.resourcesUpdates[i]; i++) {
			if (dojo.isArray(resUpdates) &&
					resUpdates[0] instanceof rmpw.common.vcs.ChangeSet) {
				// should always be the case, as we always create ChangeSet for reference creation from pop-up form
				// (see ChangeMixin.applyChange);
				// by now, as all reasources are created, we know target.rdf_about for each change, 
				// and that represents logicalUri of the resource to be updated by this chage set
				logicalUri = resUpdates[0].changes[0].target.rdf_about;
				if (!new dojo._Url(logicalUri).scheme) {
					// means this is a fragment - ignore, the change is saved with its document
					continue;
				}
				found = false;
				for (var j = 0, resUriWithUpdates; resUriWithUpdates = resourceUrisWithUpdates[j]; j++) {
					if (resUriWithUpdates.uri === logicalUri) {
						// there are allready one or more updates registered for this resource, add to it
						found = true;
						resUriWithUpdates.changeSets.push(resUpdates);
						break;
					}
				}
				if (!found) {
					// first update for this resource
					resourceUrisWithUpdates.push({"uri" : logicalUri, "changeSets" : resUpdates});
				}
			}				
		}
		// one save request per document
		for (var i = 0, resUriWithUpdates; resUriWithUpdates = resourceUrisWithUpdates[i]; i++) {		
			this.jfsStore.saveResource(resUriWithUpdates.changeSets, null, null,  
				null, true, null, resUriWithUpdates.uri);			
		}		

		this.resourcesUpdates = [];
	},
		
	isEmpty: function() {
		return this.resources.length < 1;
	}
	
});

dojo.declare("rmpw.common.store.JfsResource._Resource", null, {
	
	types: null, 
	namespaces: null,
	model: null,
	identifier: null,
	types: null,
	onResourceCreation: null,
	
	constructor: function(model, types, namespaces, identifier, target, onResourceCreation){
		this.model = model;
		this.types= types;
		this.namespaces = namespaces;
		this.identifier = identifier;
		this.target= target;
		this.onResourceCreation = onResourceCreation;
	}
});

})();