/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2011, 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmpw.diagrams.edit.tools.ActionBarScope");

dojo.require("rmpw.common.policy.Policy");
dojo.require("rmpw.diagrams.edit.actions.Redo");
dojo.require("rmpw.diagrams.edit.actions.Undo");
dojo.require("dojo.i18n");

dojo.requireLocalization("rmpw.diagrams.edit", "BehaviorMessages");
var behaviorMessages = dojo.i18n.getLocalization("rmpw.diagrams.edit", "BehaviorMessages");        

/**
 * Manages diagram toolbar. It adds static tools to it (that don't change with selection), 
 * and allows shapes to make their contributions (through actionbar policy, when the shape is selected).
 * 
 * @author Lidija Grahek
 */
dojo.declare("rmpw.diagrams.edit.tools.ActionBarScope", rmpw.common.policy.Policy, {
	singleton: "actionBarScope",
	
	constructor: function() {
		this._connections = [];
		this.selectionScope = this.findNearestParentByKey('selectionScope', true);	
		// see comment in handleSelection method
		this._connections.push(dojo.connect( this.selectionScope, "onSelectionChanged", this, "_onSelectionChanged" ));		
		this._lineStyler = new rmpw.jazz.ui.LineStyler();
		// TODO: following was taken from rmpw.dsv.editor.domain.ui.DiagramWidget and was commented there, 
		// we should make it work at some point
		// this.lineStyler.setLineColor = dojo.hitch( this, "setLineColor" ); 
		// this.lineStyler.setLineWidth = dojo.hitch( this, "setLineWidth" ); 
		// this.lineStyler.setLineStyle = dojo.hitch( this, "setLineStyle" );
		this.linetools = false;
	},
	
	setToolbarWidget: function(toolbar) {
		this.toolbar = toolbar;
	},
	
	/**
	 * Adds the given item to the toolbar. 
	 */
	addTool: function(item){		
		this.toolbar.addChild(item);
	},
	
	/**
	 * Removes the given item from the toolbar (without destroying the item).
	 */
	removeTool: function(item){
		this.toolbar.removeChild(item);
	},
	
	/**
	 * Destroys the given item.
	 */
	destroyTool: function(item){
		item.destroyRecursive();
	},
	
	/**
	 * Adds given widgets to the toolbar.
	 */
	addTools: function(widgets) {
		for (var j = 0, widget; widget = widgets[j]; j++) {			
			this.addTool(widget);
		}
	},
	
	/**
	 * Removes the widgets from the toolbar and destroys them.
	 */
	destroyTools: function(widgets) {
	    var tool, i;
		if (widgets && widgets.length > 0) {
			for (i = 0; tool = widgets[i]; i++) {
				this.destroyTool(tool);
			}
		}
	},		
	
	/**
	 * Creates a button based on given arguments and adds it to the toolbar. 
	 */
	addButton: function( label, call, icon, show ){ 
		if( !show ){ 
		    show = false; 
		}

		var button = new dijit.form.Button({ 
				label: label, 
				showLabel: show, 
				onClick: call,
				iconClass: icon,
				style: "font-size:10px"});

		this.addTool( button );

		return button;
	},	
	
	/**
	 * Adds the static toolbar widgets (that don't change with selection). 
	 * Dynamic ones (depend on the currently selected shape) are contributed by actionbar 
	 * policies installed on shapes when needed.
	 */
	addStaticToolbarWidgets: function(args) {
		this.diagramLayer = args.diagramLayer;
		var controls = rmpw.jazz.ui.Controls.items;
		var tools = args.toolInfo;
        for(var item = 0; item < tools.length; item++) {
            if(tools[item] && tools[item].display === true) {
                switch( tools[item].name ) {
                    case controls.ZOOM:
                        this._addZoom(args.rootLayer);
                        break;

                    case controls.UNDO:
                        this.addButton(behaviorMessages.undo, dojo.hitch(this, "_undo"), "undoIcon");
                        break;

                    case controls.REDO:
                        this.addButton(behaviorMessages.redo, dojo.hitch(this, "_redo"), "redoIcon");
                        break;
                }
            }
        }	
	 
	},
	
	_addZoom : function(rootLayer) {
		// code in this methd is taken from rmpw.dsv.editor.domain.ui.DiagramWidget 
		// (and was previously in rmpw.jazz.ui.Layout)
		this.slider = new rmpw.widgets.slider.HorizontalSlider({
				name: "slider",
				value: 1,
				minimum: 0.20,
				maximum: 4,
				intermediateChanges: true,
				style: "width:100px;vertical-align:middle;top:5px;",
				discreteValues: 20, // 20% - 400% at 20% intervals
				showButtons: false,
				handleSrc: dojo.moduleUrl( "rmps.graphics","images/etool16/SliderHandle.png" )
		});
 
		if ( !rmps.app.core.util.checkIfMobile() ){
			this.addTool( this.slider );
		}

		this.label = new rmps.graphics.ui.LabelWidget();
        this.label.setLabel("100%");
        this.addTool(this.label); 

		this.percent = this.addButton( "100%", dojo.hitch( this, "_resetZoom" ), "actualIcon", false );
					
		var percentLabel = this.label; 
		
		if( this.slider ){
			// catch when the slider changes the zoom
			this.slider.onChange = dojo.hitch(this, function(value) {
				rootLayer.setScale(value, null, null, true );			
				var val = Math.round( 100 * value );				
				var label = val + "%";				
				percentLabel.setLabel( label );
			});
			
			// catch when something else like the WheelZoom edit policy changes the zoom
			var viewport = this.host.root.viewer.viewport;
			this._connections.push(dojo.connect(viewport, "onViewportChanged", this, function() {
				var value = viewport.getZoom();
				var val = Math.round( 100 * value );				
				percentLabel.setLabel(val + "%");
            }));	
			this._connections.push(dojo.connect(rootLayer, "setScale", dojo.hitch(this, function(scale) {
				this.slider.set("value", scale);
			})));
		}		
	},
	
	_resetZoom: function(){
		this.slider.set( 'value', 1 );
		this.diagramLayer.setScale(1);
	},

	_redo: function() {
		var action = new rmpw.diagrams.edit.actions.Redo();
		action.run({viewer : this.host.root});
	},
		
	_undo: function() {
		var action = new rmpw.diagrams.edit.actions.Undo();
		action.run({viewer : this.host.root});
	},

	_onSelectionChanged: function() {
		// TODO: we should probably install actionbar policy on edges, and add action bindings to it 
		// based on what is in lineStyler currently (requires addition to code generation); 
		// this method can then be removed, along with _selectionIncludesLine
		var selection = this.selectionScope.getSelection();
		if (selection && selection.length > 0 && this._selectionIncludesLine(selection)) {
            if( this.linetools === false ) {
				this._lineStyler.addStyleTools(this.toolbar);
				this.linetools = true;
			}
			this._lineStyler.referenceLines(this.selection);
		} else {
			this._lineStyler.removeStyleTools();
			this.linetools = false;
		}
	},
	
	_selectionIncludesLine: function(selection) {
		// 24831: Line attribute widgets do nothing on Sketcher or any DTK editor
		// always returning false until actions are implemented
		return false;
	},
	
	/**
	 * Removes all tools from the toolbar other than those related to zoom. 
	 * Removed tools can be later re-added by calling reAddRemovedTools. 
	 */
	removeToolsExceptZoom: function() {
		this._removedTools = [];
		var i = 0;
		var toolbarItems = this.toolbar.getChildren();
		while (i < toolbarItems.length) {
			if (toolbarItems[i] !== this.slider
					&& toolbarItems[i] !== this.label 
					&& toolbarItems[i] !== this.percent) {
				// skip removal of zoom related tools
				this._removedTools.push(toolbarItems[i]);
				this.removeTool(toolbarItems[i]);
			}
			i++;
        }
	},
	
	/**
	 * Assumes that removeToolsExceptZoom was called before. 
	 * Appends all tools removed by that call back to the toolbar.
	 */
	reAddRemovedTools: function() {
		if (!this._removedTools) {
			return;
		}
		for (var i = 0; i < this._removedTools.length; i++) {
			this.addTool(this._removedTools[i]);
		}
		this._removedTools = [];
	},	
	
	destroy: function() {
		dojo.forEach(this._connections, function(conn){
			dojo.disconnect(conn);
		});
		// remove all tools from the toolbar
		this._lineStyler.removeStyleTools();
		this.label = null;
		this.slider = null;
		this.percent = null;				
		while (this.toolbar.getChildren().length > 0) {
			this.destroyTool(this.toolbar.getChildren()[0]);
		}
	}
});

}());