/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.transform.Mapping");

dojo.require("rmpw.common.policy.Host");
dojo.require("rmpw.common.transform.Trace");
dojo.require("rmpw.common.transform.Expression");
dojo.require("rmpw.common.transform.MappingState");

dojo.declare("rmpw.common.transform.Mapping", rmpw.common.policy.Host , {
		
	constructor: function(options) {
		// root elements of the semantic model
		this.sources = { __metadata: {} };
		// named elements of the view element trees
		this.template = {};
		// root elements of the the view element trees (one level below those passed in from the parent)
		this.rootShapes = [];
		
		this._options = options;
	},
	
	idToSemanticMap: {},
	
	/*
	 * Ensures that the source reference is a real resolved json object.
	 */
	getResolvedSemantic: function(source) {
		if (source && source.rdf_about) {
			var realSource = this.idToSemanticMap[source.rdf_about];
			return realSource ? realSource : source;
		}
		else {
			return source;
		}
	},
	
	/**
	 * Called when the view mapping is attached to its tree, after source and target
	 * model root element setters have been called.
	 */
	setParent: function(parent) {
		this.inherited(arguments);

		// TODO: eliminate references to this.viewer, and then the property itself
		this.viewer = this.root;
		
		this.level = parent.level + 1;		
		this.id = this.root.getInvalidationController().getNextID();
		
		this.setAll(this._options);
		delete this._options;
		
		this.createTraces();
		this.createPolicies();
		
		// set __mappings reverse references
		for(var sourceName in this.sources) {
			var source = this.sources[sourceName];
			
			if(source && dojo.isObject(source)) {
				if (source.rdf_about) {
					this.idToSemanticMap[source.rdf_about] = source;
				}
				
				if (source.__inactiveMappings) {
					for (var i=0, ms; ms=source.__inactiveMappings[i]; i++) {
						if (source.__inactiveMappings.length == 1) {
							delete source.__inactiveMappings;
						}
						else {
							source.__inactiveMappings.splice(i--, 1);
						}
						if (ms.restoreState(this) || !source.__inactiveMappings) {
							break;
						}
					}
				}
				
				if (source.__mappings) {
					source.__mappings.push(this);
				}
				else {
					source.__mappings = [this];
				}
			}
			else {
				var key = source + "";
				key =  sourceName + ":" + key;
				
				// create map from source to mapping in the parent
				if (!parent.__sourceToMapping) {
					parent.__sourceToMapping = {};
				}
				parent.__sourceToMapping[key] = this;
			}
		}
		
		this.createFixedChildren();
		this.applyModelChanges();
		this.attachListeners();
		this.invalidate();
	},
	
	createPolicies: function() {		
    	// abstract method
	},	
	
	/**
	 * Create children that only depend on parameters that remain fixed for the lifetime of the parent
	 * mapping: source roots, template variables and constants
	 */
	createFixedChildren: function() {
		// abstract method
	},
		
	/*
	 * Connect to any relevant method calls to listen to incoming events.
	 */
	attachListeners: function() {
		if(this.listeners && this.listeners.length != 0) {
			console.debug("Exception: listeners already connected");
			return;
		}
		
		this.listeners = [];
		this.listeners.push(dojo.connect(this, "applyModelChanges", this, "enforcePolicies"));
	},
	
	detachListeners: function() {
		if(!this.listeners) {
			console.warn("Listeners already disconnected");
			return;
		}
	
		dojo.forEach(this.listeners, dojo.disconnect);
		delete this.listeners;
	},
	
	destroy: function() {
		if(!this.policies) {
			console.error("Render mapping was already deleted.");
			return;
		}
		
		this.detachListeners();
		
		// remove __mapping reverse references
		if (this.sources) {
			var mappingState = new rmpw.common.transform.MappingState(this);
			
			for (var sourceName in this.sources) {
				if(sourceName.substr(0, 2) != '__') {
					var source = this.sources[sourceName];
					if(source && dojo.isObject(source)) {
						
						if (source.rdf_about) {
							delete this.idToSemanticMap[source.rdf_about];
						}
						
						if (source.__inactiveMappings) {
							source.__inactiveMappings.push(mappingState);
						}
						else {
							source.__inactiveMappings = [mappingState];
						}
						
						var list = source.__mappings;
						var found = false;
						if(list) {
							if(list.length == 1 && list[0] == this) {
								delete source.__mappings;
								found = true;
							} else {
								var i = dojo.indexOf(list, this);
								if(i != -1) {
									list.splice(i, 1);
									found = true;
								}
							}
						}
						
						if(!found) {
							console.warn("Mapping backreference in ", sourceName, " not found for ", this.declaredClass, " during destroy");
						}
					}
				}
			}
		}
		delete this.sources;
		
		// policies deleted here
		this.inherited(arguments);
		
		var rootShapes = this.rootShapes;
		for(var i = 0, shape; shape = rootShapes[i]; i++) {
			shape.getParent().remove(shape);
		}
		delete this.rootShapes;
		delete this.template;
	},
	
	invalidate: function() {
		this.viewer.getInvalidationController().invalidate(this);
	},
	
	applyModelChanges: function() {
		// process deletions, but leave additions to subclasses
		for(var traceName in this.getQueries()) {
			var trace = this.getQuery(traceName);
			if(trace.isCollection() && !trace.isSynced()) {
				if (trace.removed && trace.removed.length > 0) {
					this.processChildrenToRemove(trace, trace.removed);
				}
				else if (trace.removedReference && trace.removedReference.length > 0){
					this.processChildrenToRemoveReference(trace);
				}
			}
		}
		
		if (this._boundExpressions) {
			for(var i = 0, expressionSetter; expressionSetter = this._boundExpressions[i]; i++) {
				expressionSetter();
			}
		}
	},
		
	/*
	 * Utility method for setting an expression value into a policy method.
	 * This needs to be done through utility since the expression variable may not
	 * be named.
	 */
	bind: function(obj, setter, expression) {
		if (!this._boundExpressions) {
			this._boundExpressions = [];
		}
		
		this._boundExpressions.push(function() {
    		if (expression && !expression.isSynced()) {
    			obj[setter](expression.get());
    			expression.reset();
    		}
		});
	},
	
	/**
	 * overridden in generated domain implementations for property mappings
	 */ 
	createTraces: function() {
		this.traces = {};
		this.expressions = {};
	},
	
	/**
	 * @param {string} name of the trace variable
	 * for the rest, see Trace
	 */
	createTrace: function(name, a, b, c) {
		// TODO (MG): splice arguments instead?
		this.traces[name] = new rmpw.common.transform.Trace(this.sources, a, b, c);
		this.traces[name].__name = name;
	},
	
	/**
	 * Create an expression based on multiple trace variables
	 */
	createExpression: function(name, traceNames, impl, propertyName, collection) {
		var traces = dojo.map(traceNames, function(name) { return this.getQuery(name); }, this);
		this.expressions[name] = new rmpw.common.transform.Expression(traces, impl, propertyName, collection); 
	},
	
	/*
	 * Gets all queries associated with this mapping
	 */
	getQueries: function() {
		var queries = {};
		dojo.mixin(queries, this.traces);
		dojo.mixin(queries, this.expressions);
		return queries;
	},
	
	/*
	 * Given a name will find the associated query (typically a trace or expression)
	 * 
	 * TODO (MG): align the name of this function with QVT concepts
	 */
	getQuery: function(name) {
		return this.traces[name] || this.expressions[name];
	},
	
	getSourceByPath: function(path) {
		var getter = new Function("root", "return root." + path + ";");
		try {
			return getter(this.sources);
		} catch (e) {
			return null;
		}
	},

	getSourceOwnerByPath: function(path) {
		// TODO: can fail for array notation [], solution available in Trace.js
		var lastDot = path.lastIndexOf('.');
		if(lastDot != -1) {
			return this.getSourceByPath(path.substring(0, lastDot));
		} else {
			return null;
		}
	},

	/**
	 * Returns the first source reference, generally under the assumption
	 * there is only one.
	 * 
	 * TODO: add option to filter for type and flag to check/return
	 * 		whether there is more than one key in this.sources
	 *
	 * @returns 
	 */
	getDefaultSource: function() {
		var source;
		for(var src in this.sources) {
			if(src.substring(0, 2) != "__") { // ignore injected __graph source
				source = this.sources[src];
				break;
			}
		}
		return source;
	},

	/**
	 * Returns the source path names.
	 */
	getSourcePathNames: function() {
		var source = [];
		for(var src in this.sources) {
			if(src.substring(0, 2) != "__") { // ignore injected __graph source
				source.push(src);
			}
		}
		return source.join(",");
	},

	/**
	 * Checks to see if .
	 *
	 * @returns true if this mapping and mapping argument have the save source(s)
	 */
	hasSameModelSource: function(mapping) {
		for(var src in this.sources) {
			if(src.substring(0, 2) != "__") { // ignore injected __graph source
				if (this.sources[src] != mapping.sources[src]) {
					return false;
				}
			}
		}
		return true;
	},
	
	connectTraces: function(source, traces) {
		// validate trace
		for(var i = 0, traceName; traceName = traces[i]; i++) {
			if (this.getQuery(traceName) == null) {
				throw "Cannot find trace or expression named " + traceName;
			}
		}

		// connect trace callback
		var callback = dojo.hitch(this, function(e) {
			this._onTraceSourceChanged(traces, e);
		});
		
		if (!this.listeners) {
			this.listeners = [];
		}
		
		if (!dojo.isOffscreen) { // don't listen to events if rendering offscreen
			this.listeners.push(dojo.connect(source, "onChange", callback));	
		}
	},
	
	/**
	 * Invalidates a trace and the entire view mapping if a trace has changed, based
	 * on an event fired on the object that owns the property it pulls data from.
	 * 
	 * Usage:
	 * dojo.connect(<source object>, "onChange, dojo.hitch(this, function(e) {
	 * 		this.onTraceSourceChanged([<array of trace vars>], e);
	 * }));
	 * 
	 * @param traces list of trace variables potentially affected by the change event
	 * @param e the change event describing what changed on the object
	 * @returns
	 */
	_onTraceSourceChanged: function(traces, e) {
		var invalid = false;
		
		for(var i = 0, trace; trace = this.getQuery(traces[i]); i++) {
			if ((e.updated && trace.propertyName in e.updated) 
					|| (e.added && trace.propertyName in e.added)
					|| (e.deleted && trace.propertyName in e.deleted)) {
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.pushed && trace.propertyName in e.pushed) {
				dojo.forEach(e.pushed[trace.propertyName], function(item) {
					trace.getAdded().push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.pushedReference && trace.propertyName in e.pushedReference) {
				dojo.forEach(e.pushedReference[trace.propertyName], function(item) {
					trace.getAdded().push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.popped && trace.propertyName in e.popped) {
				dojo.forEach(e.popped[trace.propertyName], function(item) {
					trace.removed.push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.poppedReference && trace.propertyName in e.poppedReference) {
				dojo.forEach(e.poppedReference[trace.propertyName], function(item) {
					trace.removedReference.push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.insertArray && trace.propertyName in e.insertArray) {
				dojo.forEach(e.insertArray[trace.propertyName], function(item) {
					trace.getAdded().push(item.value);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.insertReferenceArray && trace.propertyName in e.insertReferenceArray) {
				dojo.forEach(e.insertReferenceArray[trace.propertyName], function(item) {
					trace.getAdded().push(item.value);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.removeArray && trace.propertyName in e.removeArray) {
				dojo.forEach(e.removeArray[trace.propertyName], function(item) {
					trace.removed.push(item.value);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.removeReferenceArray && trace.propertyName in e.removeReferenceArray) {
				dojo.forEach(e.removeReferenceArray[trace.propertyName], function(item) {
					trace.removedReference.push(item.value);
				});
				if (e.removeReferenceSourcePathNames[trace.propertyName]) {
					trace.removeReferenceSourcePathNames = e.removeReferenceSourcePathNames[trace.propertyName];
				}
				trace.setSynced(false);
				invalid = true;
			}
			// refresh
			if (e.pushedRefresh && trace.propertyName in e.pushedRefresh) {
				dojo.forEach(e.pushedRefresh[trace.propertyName], function(item) {
					trace.getAdded().push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			
		}
		
		if (invalid) {
			this.invalidate();
		}
	},
	
	addChild: function(child) {
		this.inherited(arguments);
		if (this.onChange) { // allow traces on children collection
			var e = { pushed: { children: [child] }};
			this.onChange(e);
		}
	},
	
	removeChild: function(child) {
		var removed = this.inherited(arguments);
		if (removed && this.onChange) { // allow traces on children collection
			var e = { popped: { children: [child] }};
			this.onChange(e);
		}
	},
	
	/*
	 * Deletes all associated mappings for the views being removed
	 */
	processChildrenToRemove: function(trace, views) {
		for(var i=0, lg=views.length; i<lg; i++) {
			if (dojo.isObject(views[i])) {
				if (views[i].__mappings) {
					var self = this;
					dojo.forEach(views[i].__mappings.slice(), function(mapping) {
						// only destroy this child if it was removed from this parent
						if (mapping.parent === self) {
							mapping.destroy();
						}
					});
				}
				// if __mappings is empty, mapping.destroy() deletes it on its way out
			}
			else {
				var key = views[i] + "";
				var props = trace.propertyName.split('_');
				key = props[props.length-1] + ":" + key;
				if (this.__sourceToMapping[key]) {
					this.__sourceToMapping[key].destroy();
					delete this.__sourceToMapping[key];
				}
			}
		}
	},
	
	/*
	 * Deletes the all mappings in the mapping array
	 */
	processMappingChildrenToRemove: function(mappings) {
		if (mappings) {
			dojo.forEach(mappings.slice(), function(mapping) {
				mapping.destroy();
			});
		}
	},

	/*
	 * Only removes associated mappings that are in the containment hierarchy of this mapping since
	 * the views are references (not composition in the model).
	 */
	processChildrenToRemoveReference: function(trace) {
		var views = trace.removedReference;
		var sourcePathNames = trace.removeReferenceSourcePathNames;
		for(var i=0, lg=views.length; i<lg; i++) {
			var mappings = views[i].__mappings;
			
			if (mappings) {
				var mappingsToDelete = [];
				var self = this;
				dojo.forEach(mappings, function(mapping) {
					// check that the mapping being deleted has this mapping in it's parent tree
					var parent = mapping.parent;
					while (parent != null && parent != self) {
						parent = parent.parent;
					}
					
					if (parent == self &&  (!sourcePathNames || sourcePathNames == mapping.getSourcePathNames())) {
						// delete this mapping
						mappingsToDelete.push(mapping);
					}
				});
				
				dojo.forEach(mappingsToDelete, function(mapping) {
					mappings.splice(dojo.indexOf(mappings, mapping), 1);
					mapping.destroy();
				});
				
				if (mappings.length == 0) {
					delete views[i].__mappings;
				}
			}
		}
	},
	
	getChildFromId: function(id) {
		// TODO: Should use an index instead of traversing the tree.
		if (id == this.id) {
			return this;
		}
		var child = null;
		if (this.children) {
			dojo.some(this.children, function(c) {
				if (c.id === id) {
					child = c;
					return true;
				}
				child = c.getChildFromId(id);
				if (child != null)
					return true;
				return false;
			});
		}
		return child;
	},
	
	isTypeOf: function(obj, type) {
		return this.root.commandPackage.isTypeOf(obj, type);
	},
	
	set: function(property, value) {
		// for now call legacy function naming
		this["set" + property.substring(0, 1).toUpperCase() + property.substring(1)](value);
	},
	
	setAll: function(properties) {
		for(var key in properties) {
			this.set(key, properties[key]);
		}
	}	
});