/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function(){

dojo.provide("rmps.app.ui.internal.editors.DefaultOslcContextEditor");

dojo.require("rmps.app.ui.internal.editors.BaseEditor");
dojo.require("rmps.app.ui.internal.dialogs.SelectContextDialog");

dojo.require("rmps.core.projects.ProjectClient");
dojo.require("rmps.core.util.ContextUtils");

dojo.require("jazz.ui.MessageBox");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");

var ProjectClient = rmps.core.projects.ProjectClient;
var ContextUtils = rmps.core.util.ContextUtils;

var DEFAULT_CONTEXT_PROPERTY = "defaultOslcContext";
var DIRTY_GROUP_ID = "rmps.projectProperties.DefaultOslcContextEditor.dirty";

var UA_KEY = "com.ibm.xtools.web.app.page_DefaultOslcContext001";

dojo.require("dojo.i18n");
dojo.requireLocalization("rmps.app.ui.internal.editors", "DefaultOslcContextEditorMessages");
var messages= dojo.i18n.getLocalization("rmps.app.ui.internal.editors", "DefaultOslcContextEditorMessages");

/**
 * Editor for viewing and setting the default OSLC context of a project.
 */
dojo.declare("rmps.app.ui.internal.editors.DefaultOslcContextEditor", [rmps.app.ui.internal.editors.BaseEditor], {

	isRefreshable: false, //override from superclass

	uaKey: UA_KEY,

	_context: null, //Object; the currently selected context.  Must have a "name" and "uri" field

	postCreate: function() {
		this.inherited(arguments);
		
		//load the current value of the default OSLC context
		//TODO would it be better to do all this in the editor input loading method?
		var getPropertyDeferred = ProjectClient.getProjectProperty(this.editorInput.projectId, DEFAULT_CONTEXT_PROPERTY);
		getPropertyDeferred.addCallback(this, function(propertyValue){
			//value is the URL of some stream or baseline
			if (propertyValue.length > 0) {
				this._setContext(propertyValue);
			} else {
				//treat it as property not set
				this.contentWidget.updateLabel(messages.contextNotSetLabel);
			}
		});
		getPropertyDeferred.addErrback(this, function(error) {
			if (error.status === 404) {
				//property is not set; display appropriate message
				this.contentWidget.updateLabel(messages.contextNotSetLabel);
			} else {
				this.showMessage({
					type: jazz.ui.MessageBox.STATUS_TYPES.ERROR,
					summary: messages.generalErrorMessage
				});
			}
		});		
	},
	
	/**
	 * Sets the currently selected context of this editor.
	 */
	_setContext: function(contextUrl) {
		//load the context object to get the name
		var getContextDeferred = ContextUtils.getContextDeferred(contextUrl);
		getContextDeferred.addCallback(this, function(contextObject){
			//the context object may be a Stream or a Baseline.  Both objects have a "name" field
			this._context = contextObject;
			this.contentWidget.updateLabel(contextObject.name, ContextUtils.getIconUrlForContext(contextObject.uri));
		});
		getContextDeferred.addErrback(this, function(error) {
			this.showMessage({
					type: jazz.ui.MessageBox.STATUS_TYPES.ERROR,
					summary: messages.generalErrorMessage
			});
		});
	},
	
	/**
	 * Clears the currently selected context of this editor.
	 */
	_clearContext: function() {
		this._context = null;
		this.contentWidget.updateLabel(messages.contextNotSetLabel);
	},
	
	//overridden from superclass
	createContentWidget: function() {
		return new rmps.app.ui.internal.editors._OslcContextEditorContent({projectId: this.editorInput.projectId, contextSelected: dojo.hitch(this, this.contextSelected)});
	},
	
	/**
	 *Overridden from superclass.  Saves the value of the OSLC context.
	 */
	doSave: function(taskMonitor) {
		var newValue = this._context ? this._context.uri : null;
		var deferred = ProjectClient.setProjectProperty(this.editorInput.projectId, DEFAULT_CONTEXT_PROPERTY, newValue, true);
		
		var task = taskMonitor.beginTask(messages.saveTaskLabel, 1);
		task.handleDeferredWork(deferred, 1);
	},
	
	/**
	 * Called when a new context is selected.  If contextUrl is null, the existing context should be cleared.
	 */
	contextSelected: function(contextUrl) {
		this.addDirty(this.contentWidget, DIRTY_GROUP_ID);
		
		if (contextUrl) {
			this._setContext(contextUrl);
		} else {
			this._clearContext();
		}
	}

});

dojo.declare("rmps.app.ui.internal.editors._OslcContextEditorContent", [dijit._Widget, dijit._Templated], {

	templateString: '<div>' +
						'<button style="margin-top:15px;margin-bottom:5px;" dojoAttachEvent="onclick: _select">${messages.selectButtonLabel}</button>' +
						'<label style="margin-right:5px;margin-left:5px;" dojoAttachPoint="_selectedLabel"></label>' +
						'<br><button dojoAttachEvent="onclick: _clear">${messages.clearButtonLabel}</button>' +
					'</div>',
					
	//local copy for use in template string
	messages: messages,
	
	//mixin properties
	projectId: null, //String; ID of the selected project
	contextSelected: null, //function; called when a new context is selected from the dialog, or when the context is cleared
	
	/**
	 * Method called when the user clicks on the "select" button - opens picker for selecting the
	 * default context.
	 */				
	_select: function() {
		var dialog = new rmps.app.ui.internal.dialogs.SelectContextDialog({
			projectId: this.projectId, 
			contextSelected: dojo.hitch(this, this.contextSelected)
		});
		dialog.show();
	},
	
	/**
	 * Method called when the user clicks the "clear" button - unsets the default OSLC context.
	 */
	_clear: function() {
		this.contextSelected(null);	
	},
	
	/**
	 * Updates the label with the name of the currently selected context.
	 * labelText - String, the new label text - required
	 * iconUrl - URL string for the icon; may be null
	 */
	updateLabel: function(labelText, iconUrl) {
		this._selectedLabel.innerHTML = "";
		
		if (iconUrl) {
			var img = dojo.create("img", {src:iconUrl}, this._selectedLabel);
		}
		
		this._selectedLabel.appendChild(dojo.doc.createTextNode(labelText));
	}

});

})();

